#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import sys
import six
import platform


SYSTEM_ENCODING = 'gbk' if os.name == 'nt' else 'utf-8'

# if platform.system() in ('Linux', 'Darwin') and \
#         sys.stdout.encoding and sys.stdout.encoding.upper() != 'UTF-8':
#     print("""\033[93mWarning: System "{}" python's encoding is "{}".
# Chinese may not print normally, fix with the following command
#     export PYTHONIOENCODING=UTF-8\033[0m""".format(platform.system(), sys.stdout.encoding))


def encode(s, encoding=None, errors='ignore'):
    us = s if isinstance(s, six.text_type) else decode(s)
    return us.encode(encoding or SYSTEM_ENCODING, errors)


def decode(s, encodings=['utf-8', 'gbk', 'cp936']):
    if isinstance(s, six.text_type):
        return s

    for enc in encodings:
        try:
            return six.text_type(s, enc)
        except:
            pass
    raise UnicodeDecodeError(','.join(encodings), "", 0, len(s or ''), "string: '%s'" % repr(s))


if __name__ == '__main__':
    print('Hello 世界!')
    print(encode('Hello 世界!'))
    print(encode(u'Hello 世界!'))
    print(decode('Hello 世界!'))
    print(decode(u'Hello 世界!'))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import requests


def http_download(url, target_path):
    """Download file to local
    Args:
        - url(string): url request path
        - target_path(string): download destination
    """
    r = requests.get(url, stream=True)
    with open(target_path, 'wb') as f:
        # shutil.copyfileobj(resp, f)
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    return target_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# 2014/07/02 by codeskyblue
#
'''
convert adb event to air.test program
usage is simple:
    python airtest_recorder.py output.py
'''

import sys
import time
import subprocess
import re
import os

DEVSCREEN = '/dev/input/event1'

__header__ = '''
import airtest
app = airtest.connect("{id}")
'''

def getRawShape():
    output = subprocess.check_output('adb shell getevent -p '+DEVSCREEN, shell=True)
    re.compile(r'0035.*max (\d+)')
    max_x = re.search(r'0035.*max (\d+)', output).group(1)
    max_y = re.search(r'0036.*max (\d+)', output).group(1)
    return max_x, max_y

def getShape():
    rsRE = re.compile('\s*mRestrictedScreen=\(\d+,\d+\) (?P<w>\d+)x(?P<h>\d+)')
    for line in subprocess.check_output('adb shell dumpsys window', shell=True).splitlines():
        m = rsRE.match(line)
        if m:
            return m.groups()
    raise RuntimeError('Couldn\'t find mRestrictedScreen in dumpsys')

def getScale():
    rawx, rawy = getRawShape()
    w, h = getShape()
    width, height = min(w, h), max(w, h)
    print '#raw', rawx, rawy
    print '# screen.width:', width
    print '# screen.height:', height
    if width == w:
        print("def click(x, y): app.touch(x, y)")
    else:
        print("def click(x, y): app.touch(y, %s-x)", width)
    print("def sleep(d): app.sleep(d)")
    return float(rawx)/float(width), float(rawy)/float(height)

def getDeviceId():
    output = subprocess.check_output('adb devices', shell=True)
    match = re.search(r'([\w\d:.]+)\s+device\s*$', output)
    if match:
        return match.group(1)
    raise RuntimeError("Couldn't find avaliable device")

def main(pipe, filename):
    xs, ys = [], []
    lastOper = ''
    touchStart = 0
    start = time.time()
    begin = time.time()

    deviceId = getDeviceId()
    scaleX, scaleY = getScale()

    def record(fmt, *args):
        outstr = fmt % args
        if filename:
            with open(filename, 'a') as file:
                file.write(outstr + '\n')
        print outstr

    record(__header__.format(id=deviceId))

    # plen()
    while True:
        line = pipe.readline()
        if not line.startswith(DEVSCREEN):
            continue
        channel, event, oper, value = line.split()
        # print event, oper, value#int(value, 16)
        value = int(value, 16)
        if oper == 'SYN_REPORT':
            continue
        
        if oper == 'ABS_MT_POSITION_X':
            xs.append(value)
        elif oper == 'ABS_MT_POSITION_Y':
            ys.append(value)
        elif oper == 'SYN_MT_REPORT':
            if lastOper == oper:
                xs = map(lambda x: x/scaleX, xs)
                ys = map(lambda y: y/scaleY, ys)
                if len(xs) != 0 and len(ys) != 0: # every thing is OK
                    (x1, y1), (x2, y2) = (xs[0], ys[0]), (xs[-1], ys[-1])
                    dist = ((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1))**0.5

                    duration = time.time()-touchStart
                    # print 'Duration:', duration
                    # touch up
                    if dist < 50:
                        record('app.click((%d, %d))', x1, y1)
                    else:
                        record('app.drag((%d, %d), (%d, %d))', x1, y1, x2, y2)
                xs, ys = [], []
            else:
                if len(xs) == 1:
                    # touch down
                    record('app.sleep(%.2f)', float(time.time()-start))
                    start = time.time()
                    touchStart = time.time()      
        lastOper = oper

if __name__ == '__main__':
    try:
        filename = None if len(sys.argv) == 1 else sys.argv[1]
        if filename and os.path.exists(filename):
            os.unlink(filename)
        p = subprocess.Popen(['adb', 'shell', 'getevent', '-l'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        main(p.stdout, filename)
    except KeyboardInterrupt:
        print 'Exit'
        p.kill()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# License under MIT

from __future__ import absolute_import

import collections
import copy
import functools
import inspect
import os
import sys
import time
import traceback
import warnings

import six
import cv2
import aircv as ac
from uiautomator import AutomatorDeviceObject

from atx import base
from atx import consts
from atx import errors
from atx import imutils
from atx import logutils
from atx.base import nameddict
from atx.drivers import Pattern, Bounds, FindPoint


warnings.simplefilter('default')

__dir__ = os.path.dirname(os.path.abspath(__file__))
log = logutils.getLogger(__name__)


Traceback = collections.namedtuple('Traceback', ['stack', 'exception'])
HookEvent = nameddict('HookEvent', ['flag', 'args', 'kwargs', 'retval', 'traceback', 'depth', 'is_before'])

def hook_wrap(event_type):
    def wrap(fn):
        @functools.wraps(fn)
        def _inner(*args, **kwargs):
            func_args = inspect.getcallargs(fn, *args, **kwargs)
            self = func_args.get('self')
            self._depth += 1

            def trigger(event):
                for (f, event_flag) in self._listeners:
                    if event_flag & event_type:
                        event.args = args[1:]
                        event.kwargs = kwargs
                        event.flag = event_type
                        event.depth = self._depth
                        f(event)
            
            _traceback = None
            _retval = None
            try:
                trigger(HookEvent(is_before=True))
                _retval = fn(*args, **kwargs)
                return _retval
            except Exception as e:
                _traceback = Traceback(traceback.format_exc(), e)
                raise
            finally:
                trigger(HookEvent(is_before=False, retval=_retval, traceback=_traceback))
                self._depth -= 1
        return _inner
    return wrap

class DeviceMixin(object):
    def __init__(self):
        self.image_match_method = consts.IMAGE_MATCH_METHOD_TMPL
        self.image_match_threshold = 0.8
        self._resolution = None
        self._bounds = None
        self._listeners = []
        self._depth = 0 # used for hook_wrap
        self.__last_screen = None
        self.__keep_screen = False
        self.__screensize = None

    @property
    def resolution(self):
        return self._resolution

    @resolution.setter
    def resolution(self, value):
        if value is None:
            self._resolution = None
        else:
            if not isinstance(value, tuple) or len(value) != 2:
                raise TypeError("Value should be tuple, contains two values")
            self._resolution = tuple(sorted(value))

    @property
    def last_screenshot(self):
        return self.__last_screen if self.__last_screen else self.screenshot()

    def _open_image_file(self, path):
        realpath = base.lookup_image(path, self.__screensize[0], self.__screensize[1])
        if realpath is None:
            raise IOError('file not found: {}'.format(path))
        return imutils.open(realpath)

    def pattern_open(self, image):
        if self.__screensize is None:
            self.__screensize = self.display

        if isinstance(image, Pattern):
            if image._image is None:
                image._image = self._open_image_file(image._name)
            return image
        
        if isinstance(image, six.string_types):
            path = image
            return Pattern(path, image=self._open_image_file(path))
        
        if 'numpy' in str(type(image)):
            return Pattern('unknown', image=image)
        
        raise TypeError("Not supported image type: {}".format(type(image)))

    def delay(self, secs):
        """Delay some seconds
        Args:
            secs: float seconds

        Returns:
            self
        """
        secs = int(secs)
        for i in reversed(range(secs)):
            sys.stdout.write('\r')
            sys.stdout.write("sleep %ds, left %2ds" % (secs, i+1))
            sys.stdout.flush()
            time.sleep(1)
        sys.stdout.write("\n")
        return self

    def exists(self, pattern, **match_kwargs):
        """Check if image exists in screen

        Returns:
            If exists, return FindPoint, or
            return None if result.confidence < self.image_match_threshold
        """
        ret = self.match(pattern, **match_kwargs)
        if ret is None:
            return None
        if not ret.matched:
            return None
        return ret

    def wait(self, pattern, timeout=10.0, safe=False, **match_kwargs):
        """Wait till pattern is found or time is out (default: 10s)."""
        t = time.time() + timeout
        while time.time() < t:
            ret = self.exists(pattern, **match_kwargs)
            if ret:
                return ret
            time.sleep(0.2)
        if not safe:
            raise errors.ImageNotFoundError('Not found image %s' %(pattern,))

    def wait_gone(self, pattern, timeout=10.0, safe=False, **match_kwargs):
        t = time.time() + timeout
        while time.time() < t:
            ret = self.exists(pattern, **match_kwargs)
            if not ret:
                return True
            time.sleep(0.2)
        if not safe:
            raise errors.ImageNotFoundError('Image not gone %s' %(pattern,))

    def touch(self, x, y):
        """ Alias for click """
        self.click(x, y)

    def _cal_scale(self, pattern=None):
        scale = 1.0
        resolution = (pattern and pattern.resolution) or self.resolution
        if resolution is not None:
            ow, oh = sorted(resolution)
            dw, dh = sorted(self.display)
            fw, fh = 1.0*dw/ow, 1.0*dh/oh
            # For horizontal screen, scale by Y (width)
            # For vertical screen, scale by X (height)
            scale = fw if self.rotation in (1, 3) else fh
        return scale

    @property
    def bounds(self):
        if self._bounds is None:
            return None
        return self._bounds * self._cal_scale()
    
    def _match_auto(self, screen, search_img, threshold):
        """Maybe not a good idea
        """
        # 1. try template first
        ret = ac.find_template(screen, search_img)
        if ret and ret['confidence'] > threshold:
            return FindPoint(ret['result'], ret['confidence'], consts.IMAGE_MATCH_METHOD_TMPL, matched=True)

        # 2. try sift
        ret = ac.find_sift(screen, search_img, min_match_count=10)
        if ret is None:
            return None

        matches, total = ret['confidence']
        if 1.0*matches/total > 0.5: # FIXME(ssx): sift just write here
            return FindPoint(ret['result'], ret['confidence'], consts.IMAGE_MATCH_METHOD_SIFT, matched=True)
        return None

    def match_all(self, pattern):
        """
        Test method, not suggested to use
        """
        pattern = self.pattern_open(pattern)
        search_img = pattern.image
        screen = self.region_screenshot()
        screen = imutils.from_pillow(screen)
        points = ac.find_all_template(screen, search_img, maxcnt=10)
        return points

    def match(self, pattern, screen=None, rect=None, offset=None, threshold=None, method=None):
        """Check if image position in screen

        Args:
            - pattern: Image file name or opencv image object
            - screen (PIL.Image): optional, if not None, screenshot method will be called
            - threshold (float): it depends on the image match method
            - method (string): choices on <template | sift>

        Returns:
            None or FindPoint, For example:

            FindPoint(pos=(20, 30), method='tmpl', confidence=0.801, matched=True)

            Only when confidence > self.image_match_threshold, matched will be True

        Raises:
            TypeError: when image_match_method is invalid
        """
        pattern = self.pattern_open(pattern)
        search_img = pattern.image

        pattern_scale = self._cal_scale(pattern)
        if pattern_scale != 1.0:
            search_img = cv2.resize(search_img, (0, 0), 
                fx=pattern_scale, fy=pattern_scale,
                interpolation=cv2.INTER_CUBIC)
        
        screen = screen or self.region_screenshot()
        threshold = threshold or pattern.threshold or self.image_match_threshold

        # handle offset if percent, ex (0.2, 0.8)
        dx, dy = offset or pattern.offset or (0, 0)
        dx = pattern.image.shape[1] * dx # opencv object width
        dy = pattern.image.shape[0] * dy # opencv object height
        dx, dy = int(dx*pattern_scale), int(dy*pattern_scale)

        # image match
        screen = imutils.from_pillow(screen) # convert to opencv image
        if rect and isinstance(rect, tuple) and len(rect) == 4:
            (x0, y0, x1, y1) = [v*pattern_scale for v in rect]
            (dx, dy) = dx+x0, dy+y0
            screen = imutils.crop(screen, x0, y0, x1, y1)
            #cv2.imwrite('cc.png', screen)

        match_method = method or self.image_match_method
        
        ret = None
        confidence = None
        matched = False
        position = None
        if match_method == consts.IMAGE_MATCH_METHOD_TMPL: #IMG_METHOD_TMPL
            ret = ac.find_template(screen, search_img)
            if ret is None:
                return None
            confidence = ret['confidence']
            if confidence > threshold:
                matched = True
            (x, y) = ret['result']
            position = (x+dx, y+dy) # fix by offset
        elif match_method == consts.IMAGE_MATCH_METHOD_SIFT:
            ret = ac.find_sift(screen, search_img, min_match_count=10)
            if ret is None:
                return None
            confidence = ret['confidence']
            matches, total = confidence
            if 1.0*matches/total > 0.5: # FIXME(ssx): sift just write here
                matched = True
            (x, y) = ret['result']
            position = (x+dx, y+dy) # fix by offset
        elif match_method == consts.IMAGE_MATCH_METHOD_AUTO:
            fp = self._match_auto(screen, search_img, threshold)
            if fp is None:
                return None
            (x, y) = fp.pos
            position = (x+dx, y+dy)
            return FindPoint(position, fp.confidence, fp.method, fp.matched)
        else:
            raise TypeError("Invalid image match method: %s" %(match_method,))

        (x, y) = ret['result']
        position = (x+dx, y+dy) # fix by offset
        if self.bounds:
            x, y = position
            position = (x+self.bounds.left, y+self.bounds.top)

        return FindPoint(position, confidence, match_method, matched=matched)

    def region(self, bounds):
        """Set region of the screen area
        Args:
            bounds: Bounds object

        Returns:
            A new AndroidDevice object

        Raises:
            TypeError
        """
        if not isinstance(bounds, Bounds):
            raise TypeError("region param bounds must be isinstance of Bounds")
        _d = copy.copy(self)
        _d._bounds = bounds
        return _d

    def keep_screen(self):
        """
        Freese screenshot, so all image functions will not take images, until call free_screen()
        """
        self.__last_screen = self.screenshot()
        self.__keep_screen = True
        inner_self = self

        class _C(object):
            def __enter__(self):
                pass

            def __exit__(self, type, value, traceback):
                inner_self.free_screen()

        return _C()
        
    def free_screen(self):
        """
        Unlock keep_screen()
        """
        self.__keep_screen = False
        return self

    def region_screenshot(self, filename=None):
        """Deprecated
        Take part of the screenshot
        """
        # warnings.warn("deprecated, use screenshot().crop(bounds) instead", DeprecationWarning)
        screen = self.__last_screen if self.__keep_screen else self.screenshot()
        if self.bounds:
            screen = screen.crop(self.bounds)
        if filename:
            screen.save(filename)
        return screen

    @hook_wrap(consts.EVENT_SCREENSHOT)
    def screenshot(self, filename=None):
        """
        Take screen snapshot

        Args:
            - filename: filename where save to, optional

        Returns:
            PIL.Image object

        Raises:
            TypeError, IOError
        """
        if self.__keep_screen:
            return self.__last_screen
        try:
            screen = self._take_screenshot()
        except IOError:
            # try taks screenshot again
            log.warn("warning, screenshot failed [2/1], retry again")
            screen = self._take_screenshot()
        self.__last_screen = screen
        if filename:
            save_dir = os.path.dirname(filename) or '.'
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            screen.save(filename)
        return screen

    def touch_image(self, *args, **kwargs):
        """ALias for click_image"""
        self.click_image(*args, **kwargs)

    def add_listener(self, fn, event_flags):
        """Listen event
        Args:
            - fn: function call when event happends
            - event_flags: for example
                EVENT_UIAUTO_CLICK | EVENT_UIAUTO_SWIPE

        Returns:
            None
        """
        self._listeners.append((fn, event_flags))

    def _trigger_event(self, event_flag, event):
        for (fn, flag) in self._listeners:
            if flag & event_flag:
                fn(event)

    @hook_wrap(consts.EVENT_ASSERT_EXISTS)
    def assert_exists(self, pattern, timeout=20.0, desc=None, **match_kwargs):
        """Assert if image exists
        Args:
            - image: image filename # not support pattern for now
            - timeout (float): seconds

        Returns:
            Find point

        Raises:
            AssertExistsError
        """
        warnings.warn("deprecated, use rp.assert_exists instead", DeprecationWarning)
        log.info('assert exists image(%s): %s', desc or '', pattern)
        start_time = time.time()
        while time.time() - start_time < timeout:
            point = self.match(pattern, **match_kwargs)
            if point is None:
                sys.stdout.write('.')
                sys.stdout.flush()
                continue
            if not point.matched:
                log.debug('Ignore confidence: %s', point.confidence)
                continue
            log.info('assert pass, confidence: %s', point.confidence)
            sys.stdout.write('\n')
            return point
        else:
            sys.stdout.write('\n')
            raise errors.AssertExistsError('image not found %s' %(pattern,))
            
    @hook_wrap(consts.EVENT_CLICK_IMAGE)
    def click_nowait(self, pattern, action='click', desc=None, **match_kwargs):
        """ Return immediately if no image found

        Args:
            - pattern (str or Pattern): filename or an opencv image object.
            - action (str): click or long_click

        Returns:
            Click point or None
        """
        point = self.match(pattern, **match_kwargs)
        if not point or not point.matched:
            return None

        func = getattr(self, action)
        func(*point.pos)
        return point

    @hook_wrap(consts.EVENT_CLICK_IMAGE)
    def click_image(self, pattern, timeout=20.0, action='click', safe=False, desc=None, delay=None, **match_kwargs):
        """Simulate click according image position

        Args:
            - pattern (str or Pattern): filename or an opencv image object.
            - timeout (float): if image not found during this time, ImageNotFoundError will raise.
            - action (str): click or long_click
            - safe (bool): if safe is True, Exception will not raise and return None instead.
            - method (str): image match method, choice of <template|sift>
            - delay (float): wait for a moment then perform click

        Returns:
            None

        Raises:
            ImageNotFoundError: An error occured when img not found in current screen.
        """
        pattern = self.pattern_open(pattern)
        log.info('click image:%s %s', desc or '', pattern)
        start_time = time.time()
        found = False
        point = None
        while time.time() - start_time < timeout:
            point = self.match(pattern, **match_kwargs)
            if point is None:
                sys.stdout.write('.')
                sys.stdout.flush()
                continue

            log.debug('confidence: %s', point.confidence)
            if not point.matched:
                log.info('Ignore confidence: %s', point.confidence)
                continue
            
            # wait for program ready
            if delay and delay > 0:
                self.delay(delay)

            func = getattr(self, action)
            func(*point.pos)

            found = True
            break
        sys.stdout.write('\n')

        if not found:
            if safe:
                log.info("Image(%s) not found, safe=True, skip", pattern)
                return None
            raise errors.ImageNotFoundError('Not found image %s' % pattern, point)

        # FIXME(ssx): maybe this function is too complex
        return point #collections.namedtuple('X', ['pattern', 'point'])(pattern, point)


if __name__ == '__main__':
    b = Bounds(1, 2, 3, 4)
    print(b)
    print(b * 1.0)




import time

from airtest import patch

def test_go():
    d = {}
    @patch.go
    def echo_hello(d):
        time.sleep(0.02)
        d['name'] = 'nice'
    p = echo_hello(d)
    assert d.get('name') == None
    p.join()
    assert d.get('name') == 'nice'




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# extension for https://sites.google.com/a/chromium.org/chromedriver/
# Experimental, maybe change in the future
# Created by <hzsunshx> 2017-01-20

from __future__ import absolute_import


import atexit
import six
from selenium import webdriver

if six.PY3:
    import subprocess
    from urllib.error import URLError
else:
    from urllib2 import URLError
    import subprocess32 as subprocess


class ChromeDriver(object):
    def __init__(self, d, port=9515):
        self._d = d
        self._port = port

    def _launch_webdriver(self):
        print("start chromedriver instance")
        p = subprocess.Popen(['chromedriver', '--port='+str(self._port)])
        try:
            p.wait(timeout=2.0)
            return False
        except subprocess.TimeoutExpired:
            return True

    def driver(self, package=None, attach=True, activity=None, process=None):
        """
        Args:
            - package(string): default current running app
            - attach(bool): default true, Attach to an already-running app instead of launching the app with a clear data directory
            - activity(string): Name of the Activity hosting the WebView.
            - process(string): Process name of the Activity hosting the WebView (as given by ps). 
                If not given, the process name is assumed to be the same as androidPackage.

        Returns:
            selenium driver
        """
        app = self._d.current_app()
        capabilities = {
            'chromeOptions': {
                'androidDeviceSerial': self._d.serial,
                'androidPackage': package or app.package,
                'androidUseRunningApp': attach,
                'androidProcess': process or app.package,
                'androidActivity': activity or app.activity,
            }
        }

        try:
            dr = webdriver.Remote('http://localhost:%d' % self._port, capabilities)
        except URLError:
            self._launch_webdriver()
            dr = webdriver.Remote('http://localhost:%d' % self._port, capabilities)
        
        # always quit driver when done
        atexit.register(dr.quit)
        return dr

    def windows_kill(self):
        subprocess.call(['taskkill', '/F', '/IM', 'chromedriver.exe', '/T'])


if __name__ == '__main__':
    import atx
    d = atx.connect()
    driver = ChromeDriver(d).driver()
    elem = driver.find_element_by_link_text(u"登录")
    elem.click()
    driver.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pytest
import string

import cv2
from atx import imutils


def test_mark_point():
    im = cv2.imread('media/system-app.png')
    im = imutils.mark_point(im, 50, 50)
    cv2.imwrite('tmp.png', im)


try:
    from Tkinter import *
    import Tkinter as ttk
except:
    from tkinter import *
    from tkinter import ttk

def calculate(*args):
    try:
        value = float(feet.get())
        meters.set((0.3048 * value * 10000.0 + 0.5)/10000.0)
    except ValueError:
        pass
    
root = Tk()
root.title("Feet to Meters")

mainframe = ttk.Frame(root)#, padding="3 3 12 12")
mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
mainframe.columnconfigure(0, weight=1)
mainframe.rowconfigure(0, weight=1)

feet = StringVar()
meters = StringVar()

feet_entry = ttk.Entry(mainframe, width=7, textvariable=feet)
feet_entry.grid(column=2, row=1, sticky=(W, E))

ttk.Label(mainframe, textvariable=meters).grid(column=2, row=2, sticky=(W, E))
ttk.Button(mainframe, text="Calculate", command=calculate).grid(column=3, row=3, sticky=W)

ttk.Label(mainframe, text="feet").grid(column=3, row=1, sticky=W)
ttk.Label(mainframe, text="is equivalent to").grid(column=1, row=2, sticky=E)
ttk.Label(mainframe, text="meters").grid(column=3, row=2, sticky=W)

for child in mainframe.winfo_children(): child.grid_configure(padx=5, pady=5)

feet_entry.focus()
root.bind('<Return>', calculate)

root.mainloop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import functools
import os
import sys
import subprocess32 as subprocess
import tempfile
import inspect
import plistlib

from PIL import Image
from atx import logutils


logger = logutils.getLogger(__name__)

def look_path(name, search_paths=[], env_path=True):
    os.pathsep
    if env_path:
        search_paths += os.getenv('PATH').split(os.pathsep)
    for directory in search_paths:
        if not os.path.isdir(directory):
            continue
        filepath = os.path.join(directory, name)
        if os.path.isfile(filepath):
            return filepath
    return None


__execpath = {}

def look_exec(name):
    if __execpath.get(name):
        return __execpath[name]

    ext = '.exe' if os.name == 'nt' else ''
    search_paths = [
        r'C:\Program Files (x86)\Quamotion\iMobileDevice',
        r'D:\Program Files (x86)\Quamotion\iMobileDevice',
        r'E:\Program Files (x86)\Quamotion\iMobileDevice',
        r'F:\Program Files (x86)\Quamotion\iMobileDevice',
        r'C:\Program Files\Quamotion\iMobileDevice',
        r'D:\Program Files\Quamotion\iMobileDevice',
        r'E:\Program Files\Quamotion\iMobileDevice',
        r'F:\Program Files\Quamotion\iMobileDevice',
    ]
    filepath = look_path(name+ext, search_paths)
    __execpath[name] = filepath
    return filepath


def must_look_exec(name):
    filepath = look_exec(name)
    if not filepath:
        raise EnvironmentError('No "%s" found.' % filepath)
    return filepath


def check_output(cmds, shell=False):
    try:
        output = subprocess.check_output(cmds, stderr=subprocess.STDOUT, shell=shell)
        return output
    except subprocess.CalledProcessError:
        # logger.warn('Failed to run command: %s', ' '.join(cmds))
        # logger.warn('Error output:\n%s', e.output)
        raise


def idevice(name, *args):
    exec_name = 'idevice' + name
    exec_path = look_exec(exec_name)
    if not exec_path:
        raise EnvironmentError('Necessary binary ("%s") not found.' % exec_name)

    cmds = [exec_path] + list(args)
    try:
        output = subprocess.check_output(cmds, stderr=subprocess.STDOUT, shell=False)
        return output
    except subprocess.CalledProcessError:
        raise

def devices():
    """
    Return device dict
    For example:
    {
        "1002038889199992134bad1234112312": "Tony's iPhone"
    }
    """
    udids = [udid.strip() for udid in idevice('_id', '-l').splitlines() if udid.strip()]
    return {udid: idevice('name', '-u', udid).decode('utf-8').strip() for udid in udids}

# def device_product_version(udid):
#     return idevice('info', '-u', udid, '-k', 'ProductVersion').strip()


def memory_last(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        func_args = inspect.getcallargs(fn, *args, **kwargs)
        self = func_args.get('self')
        store_name = '__'+fn.__name__
        
        if hasattr(self, store_name):
            return getattr(self, store_name)
        ret = fn(*args, **kwargs)
        setattr(self, store_name, ret)
        return ret
    return inner


class Device(object):
    def __init__(self, udid=None):
        if not udid:
            devs = devices()
            if len(devs) == 0:
                raise EnvironmentError("No iOS devices connected.")
            elif len(devs) > 1:
                raise EnvironmentError("More than one device connected, need to specify udid")
            else:
                udid = devs.keys()[0]
        self._udid = udid
        self.__product_version = None

    @property
    def udid(self):
        return self._udid

    @property
    @memory_last
    def product_version(self):
        return idevice('info', '-u', self.udid, '-k', 'ProductVersion').strip()

    @property
    @memory_last
    def name(self):
        return idevice('name', '-u', self.udid).decode('utf-8').strip()

    @property
    @memory_last
    def info(self):
        return plistlib.readPlistFromString(idevice('info', '--xml', '--udid', self.udid))

    def screenshot(self, filename=None):
        """
        Return:
            PIL.Image
            
        Raises:
            EnvironmentError
        """
        tmpfile = tempfile.mktemp(prefix='atx-screencap-', suffix='.tiff')
        try:
            idevice("screenshot", "--udid", self.udid, tmpfile)
        except subprocess.CalledProcessError as e:
            sys.exit(e.message)

        try:
            image = Image.open(tmpfile)
            image.load()
            if filename:
                image.save(filename)
            return image
        finally:
            if os.path.exists(tmpfile):
                os.unlink(tmpfile)

    def start_app(self, bundle_id):
        '''
        Start app by bundle_id
        Args:
            - bundle_id(string): ex com.netease.my
        Returns:
            idevicedebug subprocess instance
        '''
        idevicedebug = must_look_exec('idevicedebug')

        # run in background
        kwargs = {'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}
        if sys.platform != 'darwin':
            kwargs['close_fds'] = True
        return subprocess.Popen([idevicedebug, "--udid", self.udid, 'run', bundle_id], **kwargs)

    def install(self, filepath):
        ''' TODO(ssx): not tested. '''
        if not os.path.exists(filepath):
            raise EnvironmentError('file "%s" not exists.' % filepath)

        ideviceinstaller = must_look_exec('ideviceinstaller')
        os.system(subprocess.list2cmdline([ideviceinstaller, '-u', self.udid, '-i', filepath]))


if __name__ == '__main__':
    devices()
    dev = Device()
    print dev.screenshot('i.png')
    print dev.product_version
    print dev.product_version
    print dev.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# USAGE
# python -matx -s ESLKJXX gui

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import argparse
import functools
import json
import sys
import inspect
from contextlib import contextmanager

from atx.cmds import run
from atx.cmds import iosdeveloper
from atx.cmds import install


def inject(func, kwargs):
    '''
    This is a black techonology of python. like golang martini inject

    Usage:
    allargs = dict(name='foo', message='body')
    def func(name):
        print 'hi', name

    inject(func, allargs)
    # will output
    # hi foo
    '''
    args = []
    for name in inspect.getargspec(func).args:
        args.append(kwargs.get(name))
    return func(*args)


def load_main(module_name):
    def _inner(parser_args):
        module_path = 'atx.cmds.'+module_name
        __import__(module_path)
        mod = sys.modules[module_path]
        pargs = vars(parser_args)
        return inject(mod.main, pargs)
    return _inner


def _apk_parse(args):
    import atx.androaxml as apkparse
    (pkg_name, activity) = apkparse.parse_apk(args.filename)
    print(json.dumps({
        'package_name': pkg_name,
        'main_activity': activity,
    }, indent=4))


def _version(args):
    import atx
    print(atx.version)


def _deprecated(args):
    print('Deprecated')

def main():
    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument("-s", "--serial", "--udid", required=False, help="Android serial or iOS unid")
    ap.add_argument("-H", "--host", required=False, default='127.0.0.1', help="Adb host")
    ap.add_argument("-P", "--port", required=False, type=int, default=5037, help="Adb port")

    subp = ap.add_subparsers()

    @contextmanager
    def add_parser(name):
        yield subp.add_parser(name, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    with add_parser('tcpproxy') as p:
        p.description = 'A very simple tcp proxy'
        p.add_argument('-l', '--listen', default=5555, type=int, help='Listen port')
        p.add_argument('-f', '--forward', default=26944, type=int, help='Forwarded port')
        p.add_argument('--host', default='127.0.0.1', type=str, help='Forwarded host')
        p.set_defaults(func=load_main('tcpproxy'))

    with add_parser('gui') as p:
        p.description = 'GUI tool to help write test script'
        p.add_argument('-p', '--platform', default='auto', choices=('auto', 'android', 'ios'), help='platform')
        p.add_argument('-s', '--serial', default=None, type=str, help='android serial or WDA device url')
        p.add_argument('--scale', default=0.5, type=float, help='scale size')
        p.set_defaults(func=load_main('tkgui'))

    # Remove because of unstable
    # with add_parser('record') as p:
    #     p.add_argument('-d', '--workdir', default='.', help='workdir where case & frame files are saved.')
    #     p.add_argument('-e', '--edit', action='store_true', dest='edit_mode', help='edit old records.')
    #     p.add_argument('-a', '--nonui-activity', action='append', dest='nonui_activities',
    #         required=False, help='nonui-activities for which the recorder will analyze screen image instead of uixml.')
    #     p.set_defaults(func=load_main('record'))

    with add_parser('minicap') as p:
        p.description = 'install minicap to phone'
        p.set_defaults(func=load_main('minicap'))

    with add_parser('apkparse') as p:
        p.description = 'parse package-name and main-activity from apk'
        p.add_argument('filename', help='Apk filename')
        p.set_defaults(func=_apk_parse)

    with add_parser('monkey') as p:
        p.set_defaults(func=load_main('monkey'))

    with add_parser('install') as p:
        p.description = 'install apk to phone'
        p.add_argument('path', help='<apk file path | apk url path> (only support android for now)')
        p.add_argument('--start', action='store_true', help='Start app when app success installed')
        p.set_defaults(func=load_main('install'))

    with add_parser('screen') as p:
        p.add_argument('--scale', required=False, type=float, default=0.5, help='image scale')
        p.add_argument('--simple', action='store_true', help='disable interact controls')
        p.set_defaults(func=load_main('screen'))

    with add_parser('screencap') as p:
        p.description = 'take screenshot'
        p.add_argument('--scale', required=False, type=float, default=1.0, help='image scale')
        p.add_argument('-o', '--out', required=False, default='screenshot.png', help='output path')
        p.add_argument('-m', '--method', required=False, default='minicap', choices=('minicap', 'screencap'), help='screenshot method')
        p.set_defaults(func=load_main('screencap'))

    with add_parser('screenrecord') as p:
        p.description = 'record video (require minicap)'
        p.add_argument('-o', '--output', default='out.avi', help='video output path')
        p.add_argument('--overwrite', action='store_true', help='overwrite video output file.')
        p.add_argument('--scale', type=float, default=0.5, help='image scale for video')
        p.add_argument('-q', '--quiet', dest='verbose', action='store_false', help='display screen while recording.')
        p.add_argument('--portrait', action='store_true', help='set video framesize to portrait instead of landscape.')
        p.set_defaults(func=load_main('screenrecord'))

    with add_parser('web') as p:
        p.description = 'not maintained this func, try with: pip install atx-webide'
        p.set_defaults(func=_deprecated)

    with add_parser('run') as p:
        p.add_argument('-f', dest='config_file', default='atx.yml', help='config file')
        p.set_defaults(func=load_main('run'))

    with add_parser('version') as p:
        p.set_defaults(func=_version)

    with add_parser('info') as p:
        p.set_defaults(func=load_main('info'))

    with add_parser('doctor') as p:
        p.set_defaults(func=load_main('doctor'))

    args = ap.parse_args()
    if not hasattr(args, 'func'):
        print(' '.join(sys.argv) + ' -h for more help')
        return
    args.func(args)

if __name__ == '__main__':
    main()



#-*- encoding: utf-8 -*-
# python client for openstf STFService & Agent.
#
# Api:
#   start(adbprefix=None, service_port=1100, agent_port=1090)
#   stop(adbprefix=None, service_port=1100, agent_port=1090)
#
#   wake()              # return None
#   type(text)          # return None
#   ascii_type(text)                    # return None
#   keyevent(key, event='PRESS')        # return None
#   keyboard(char, holdtime=None)       # return None
#
#   identify()                          # return bool(success or not)
#   set_rotation(rotation, lock=False)  # return bool(success or not)
#   set_wifi_enabled(True)              # return bool(success or not)
#   get_wifi_status()                   # return a dict
#   get_display()                       # return a dict
#   get_properties(*args)               # return a dict
#
#   on_battery_event(callback)      # callback accept a dict argument
#   on_rotation_event(callback)     # callback accept a int argument
#   on_connectivity_event(callback) # callback accept a dict argument

import time
import subprocess
import select
import socket
import traceback
import Queue
import threading
import warnings

#from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.internal import encoder, decoder

import stfwire_pb2 as wire
import keycode


__all__ = ['start', 'stop', 'isalive', 'identify', 'type', 'keyevent', 'get_wifi_status', 'set_wifi_enabled', 
    'set_rotation', 'get_display', 'get_properties', 'on_battery_event', 'on_rotation_event']

messages = {
    wire.DO_IDENTIFY : (wire.DoIdentifyRequest, wire.DoIdentifyResponse),
    wire.DO_KEYEVENT : (wire.KeyEventRequest, None),
    wire.DO_TYPE: (wire.DoTypeRequest, None),
    wire.DO_WAKE: (wire.DoWakeRequest, None),
    wire.DO_ADD_ACCOUNT_MENU: (wire.DoAddAccountMenuRequest, wire.DoAddAccountMenuResponse),
    wire.DO_REMOVE_ACCOUNT: (wire.DoRemoveAccountRequest, wire.DoRemoveAccountResponse),
    wire.GET_ACCOUNTS: (wire.GetAccountsRequest, wire.GetAccountsResponse),
    wire.GET_BROWSERS: (wire.GetBrowsersRequest, wire.GetBrowsersResponse),
    wire.GET_CLIPBOARD: (wire.GetClipboardRequest, wire.GetClipboardResponse),
    wire.GET_DISPLAY: (wire.GetDisplayRequest, wire.GetDisplayResponse),
    wire.GET_PROPERTIES: (wire.GetPropertiesRequest, wire.GetPropertiesResponse),
    wire.GET_RINGER_MODE: (wire.GetRingerModeRequest, wire.GetRingerModeResponse),
    wire.GET_SD_STATUS: (wire.GetSdStatusRequest, wire.GetSdStatusResponse),
    wire.GET_VERSION: (wire.GetVersionRequest, wire.GetVersionResponse),
    wire.GET_WIFI_STATUS: (wire.GetWifiStatusRequest, wire.GetWifiStatusResponse),
    wire.SET_CLIPBOARD: (wire.SetClipboardRequest, wire.SetClipboardResponse),
    wire.SET_KEYGUARD_STATE: (wire.SetKeyguardStateRequest, wire.SetKeyguardStateResponse),
    wire.SET_RINGER_MODE: (wire.SetRingerModeRequest, wire.SetRingerModeResponse),
    wire.SET_ROTATION: (wire.SetRotationRequest, None),
    wire.SET_WAKE_LOCK: (wire.SetWakeLockRequest, wire.SetWakeLockResponse),
    wire.SET_WIFI_ENABLED: (wire.SetWifiEnabledRequest, wire.SetWifiEnabledResponse),
    wire.SET_MASTER_MUTE: (wire.SetMasterMuteRequest, wire.SetMasterMuteResponse),
    wire.EVENT_AIRPLANE_MODE: (None, wire.AirplaneModeEvent),
    wire.EVENT_BATTERY: (None, wire.BatteryEvent),
    wire.EVENT_CONNECTIVITY: (None, wire.ConnectivityEvent),
    wire.EVENT_PHONE_STATE: (None, wire.PhoneStateEvent),
    wire.EVENT_ROTATION: (None, wire.RotationEvent),
    wire.EVENT_BROWSER_PACKAGE: (None, wire.BrowserPackageEvent),
}

def pack(mtype, request, rid=None):
    '''pack request to delimited data'''
    envelope = wire.Envelope() 
    if rid is not None:
        envelope.id = rid
    envelope.type = mtype
    envelope.message = request.SerializeToString()
    data = envelope.SerializeToString()
    data = encoder._VarintBytes(len(data)) + data
    return data

def unpack(data):
    '''unpack from delimited data'''
    size, position = decoder._DecodeVarint(data, 0)
    envelope = wire.Envelope() 
    envelope.ParseFromString(data[position:position+size])
    return envelope

def _id_generator():
    request_id = [0]
    def _next():
        if request_id[0] == 0xffffffff: #uint32
            request_id[0] = 0
        request_id[0] += 1
        return request_id[0]
    return _next

get_request_id = _id_generator()

eventhooks = {}     # called when events arrived
# initialize eventhooks
for mtype, (req_class, _) in messages.iteritems():
    if req_class is None:
        eventhooks[mtype] = []
del mtype, req_class

responses = {}      # save service call response, for synchronize
service_response_lock = threading.Lock()

def route(envelope):
    _, resp_class = messages[envelope.type]
    resp = resp_class()
    resp.ParseFromString(envelope.message)
    # service calls should have id
    if envelope.id:
        with service_response_lock:
            rid = envelope.id
            # remove placeholder if the response is not needed
            if rid in responses:
                del responses[rid]
            else:
                responses[rid] = resp
        return
    # handle events
    hooks = eventhooks.get(envelope.type)
    if hooks is not None:
        for func in hooks:
            try:
                func(resp)
            except:
                traceback.print_exc()
        return
    # no handler found
    warnings.warn('No handler found for %s(%s)' % (resp_class.DESCRIPTOR.name, envelope.type))

def wait_response(rid, timeout=1):
    timeout = time.time() + timeout
    while timeout > time.time():
        with service_response_lock:
            if rid in responses:
                return responses.pop(rid)
        time.sleep(0.1)
    # cleanup timeouted response, avoid memory leak by add a placeholder!
    with service_response_lock:
        if rid in responses:
            del responses[rid]
        else:
            responses[rid] = None

def register_eventhook(mtype, callback):
    global eventhooks
    eventhooks[mtype].append(callback)

def start_stf_service(adbprefix=None, port=1100):
    if adbprefix is None: 
        adbprefix = ['adb']
    cmds = [['shell', 'am', 'startservice'],
            ['--user', '0'],
            ['-a', 'jp.co.cyberagent.stf.ACTION_START', '-n', 'jp.co.cyberagent.stf/.Service']]
    
    command = adbprefix + cmds[0] + cmds[1] + cmds[2]

    print subprocess.check_output(command)
    # if falied, using:
    # command = cmds[0] + cmds[2]
    command = adbprefix + ['forward', 'tcp:%s' % port, 'tcp:1100'] # remote port use default 1100, although it can be changed
    subprocess.call(command)

def stop_stf_service(adbprefix=None, port=1100):
    if adbprefix is None: 
        adbprefix = ['adb']
    command = adbprefix + ['forward', '--remove', 'tcp:%s' % port]
    subprocess.call(command)

    command = adbprefix + ['shell', 'am', 'stopservice', '-n', 'jp.co.cyberagent.stf/.Service']
    subprocess.call(command)

def check_stf_agent(adbprefix=None, kill=False):
    '''return True if agent is alive.'''
    if adbprefix is None: 
        adbprefix = ['adb']
    command = adbprefix + ['shell', 'ps']
    out = subprocess.check_output(command).strip()
    out = out.splitlines()
    if len(out) > 1:
        first, out = out[0], out[1:]
        idx = first.split().index('PID')
        pid = None
        for line in out:
            if 'stf.agent' in line:
                pid = line.split()[idx]
                print 'stf.agent is running, pid is', pid
                break
        if pid is not None:
            if kill:
                print 'killing', pid
                command = adbprefix + ['shell', 'kill', '-9', pid]
                subprocess.call(command)
                return False
            return True
    return False

def start_stf_agent(adbprefix=None, restart=False, port=1090):
    if adbprefix is None: 
        adbprefix = ['adb']
    if check_stf_agent(adbprefix, kill=restart):
        return

    command = adbprefix + ['shell', 'pm', 'path', 'jp.co.cyberagent.stf']
    out = subprocess.check_output(command).strip()
    path = out.split(':')[-1]
    print 'stf agent path', repr(path)
    
    command = adbprefix + ['shell', 'CLASSPATH="%s"' % path, 
            'app_process', '/system/bin', 'jp.co.cyberagent.stf.Agent']
    p = subprocess.Popen(command, stdout=subprocess.PIPE)
    # IMPORTANT! wait for agent to start listening.
    p.stdout.readline()

    command = adbprefix + ['forward', 'tcp:%s' % port, 'tcp:1090'] # remote port is 1090, cannot change
    subprocess.call(command)

def stop_stf_agent(adbprefix=None, port=1090):
    if adbprefix is None: 
        adbprefix = ['adb']
    command = adbprefix + ['forward', '--remove', 'tcp:%s' % port]
    subprocess.call(command)

    check_stf_agent(adbprefix, kill=True)

service_queue = Queue.Queue()
agent_queue = Queue.Queue()
stop_event = threading.Event()

def isalive():
    return stop_event.isSet()

def listen_service(service_port=1100): 
    # service, send & recv, use timeout
    def _service():
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(('127.0.0.1', service_port))
            while not stop_event.isSet():
                r, w, _ = select.select([s], [s], [], 0.1) 
                if s in r:
                    data = s.recv(1024)
                    if not data:
                        continue
                    try:
                        envelope = unpack(data)
                        route(envelope)
                    except:
                        print 'error while handle response'
                        traceback.print_exc()
                if s in w:
                    try:
                        message = service_queue.get_nowait()
                    except Queue.Empty:
                        pass
                    else:
                        s.sendall(message)
        except:
            traceback.print_exc()
        finally:
            s.close()
            print 'Service socket closed'
            stop_event.set()

    t = threading.Thread(target=_service)
    t.setDaemon(True)
    t.start()

def listen_agent(agent_port=1090):
    # just send, no recv
    def _agent():
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(('127.0.0.1', agent_port))
            while not stop_event.isSet():
                message = agent_queue.get() 
                s.sendall(message)
        except:
            traceback.print_exc()
        finally:
            s.close()
            print 'Agent socket closed.'
            stop_event.set()

    t = threading.Thread(target=_agent)
    t.setDaemon(True)
    t.start()

#--------------------------- API functions ----------------------------

def start(adbprefix=None, service_port=1100, agent_port=1090):
    start_stf_service(adbprefix, port=service_port)
    start_stf_agent(adbprefix, port=agent_port)
    stop_event.clear()
    listen_service(service_port)
    listen_agent(agent_port)

def stop(adbprefix=None, service_port=1100, agent_port=1090):
    stop_event.set()
    stop_stf_service(adbprefix, port=service_port)
    stop_stf_agent(adbprefix, port=agent_port)

# agent 
def keyevent(key, event=wire.PRESS, shift=False, ctrl=False, alt=False, meta=False, 
        sym=False, function=False, capslock=False, scrolllock=False, numlock=False):
    if isinstance(event, basestring):
        event = event.upper()
        assert event in ('DOWN', 'UP', 'PRESS')
        event = getattr(wire, event)
    if isinstance(key, basestring):
        if not key.startswith('KEYCODE_'):
            key = 'KEYCODE_%s' % key
        key = getattr(keycode, key)
    req = wire.KeyEventRequest()
    req.event = event
    req.keyCode = key
    req.shiftKey = shift
    req.ctrlKey = ctrl
    req.altKey = alt
    req.metaKey = meta
    req.symKey = sym
    req.functionKey = function
    req.capsLockKey = capslock
    req.scrollLockKey = scrolllock
    req.numLockKey = numlock
    msg = pack(wire.DO_KEYEVENT, req)
    agent_queue.put(msg)

# simple keyboard input
def keyboard(char, holdtime=None):
    shift = False
    ctrl = False
    if len(char) > 1:
        char = char.upper()
        if not char.startswith('KEYCODE_'):
            char = 'KEYCODE_%s' % char
        code = getattr(keycode, char, None)
        if code is None:
            print 'invalid keycode', char
            return
    elif char in 'abcdefghijklmnopqrstuvwxyz1234567890':
        code = getattr(keycode, 'KEYCODE_%s' % char.upper())
    elif char in 'ABCDEFGHIJLKMNOPQRSTUVWXYZ':
        code = getattr(keycode, 'KEYCODE_%s' % char)
        shift = True
    elif char in keycode.SHIFTED_KEYS:
        code = keycode.SHIFTED_KEYS[char]
        shift = True
    elif char in keycode.KEYBOARD_KEYS:
        code = keycode.KEYBOARD_KEYS[char]
    elif char in keycode.CTRLED_KEYS:
        code = keycode.CTRLED_KEYS[char]
        ctrl = True
    else:
        print 'invalid char', repr(char)
        return

    if holdtime is None:
        req = wire.KeyEventRequest(event=wire.PRESS, keyCode=code, shiftKey=shift, ctrlKey=ctrl)
        msg = pack(wire.DO_KEYEVENT, req)
        agent_queue.put(msg)
    else:
        # keydown
        req = wire.KeyEventRequest(event=wire.DOWN, keyCode=code, shiftKey=shift, ctrlKey=ctrl)
        msg = pack(wire.DO_KEYEVENT, req)
        agent_queue.put(msg)
        # wait
        time.sleep(holdtime)
        # keyup
        req = wire.KeyEventRequest(event=wire.UP, keyCode=code, shiftKey=shift, ctrlKey=ctrl)
        msg = pack(wire.DO_KEYEVENT, req)
        agent_queue.put(msg)

# agent, can input asciis only
def ascii_type(text):
    msg = pack(wire.DO_TYPE, wire.DoTypeRequest(text=text))
    agent_queue.put(msg)

# agent
def set_rotation(rotation, lock=False):
    msg = pack(wire.SET_ROTATION, wire.SetRotationRequest(rotation=rotation, lock=lock))
    agent_queue.put(msg)

# agent
def wake():
    msg = pack(wire.DO_WAKE, wire.DoWakeRequest())
    agent_queue.put(msg)

# use both agent & service to input unicode characters
# TODO make it atomic, maybe use another queue?
def type(text):
    rid = get_request_id()
    msg = pack(wire.SET_CLIPBOARD, wire.SetClipboardRequest(type=wire.TEXT, text=text), rid)
    service_queue.put(msg)
    # wait for clipboard result
    resp = wait_response(rid)
    if resp and resp.success:
        keyevent(keycode.KEYCODE_V, wire.PRESS, ctrl=True)
        return True
    return False

def identify():
    rid = get_request_id()
    serial = ""
    msg = pack(wire.DO_IDENTIFY, wire.DoIdentifyRequest(serial=serial), rid)
    service_queue.put(msg)
    resp = wait_response(rid)
    return resp and resp.success

def get_wifi_status():
    # return True if wifi is enabled.
    rid = get_request_id()
    msg = pack(wire.GET_WIFI_STATUS, wire.GetWifiStatusRequest(), rid)
    service_queue.put(msg)
    resp = wait_response(rid)
    return resp and resp.success and resp.status

def set_wifi_enabled(enabled):
    rid = get_request_id()
    msg = pack(wire.SET_WIFI_ENABLED, wire.SetWifiEnabledRequest(enabled=bool(enabled)), rid)
    service_queue.put(msg)
    resp = wait_response(rid, timeout=5) # may ask for user permission
    return resp and resp.success

def get_display(deviceid=0):
    rid = get_request_id()
    msg = pack(wire.GET_DISPLAY, wire.GetDisplayRequest(id=deviceid), rid)
    service_queue.put(msg)
    resp = wait_response(rid)
    if not resp or not resp.success:
        return {}
    fields = [f.name for f in wire.GetDisplayResponse.DESCRIPTOR.fields]
    data = dict([(f, getattr(resp, f)) for f in fields])
    return data

def get_properties(*args):
    rid = get_request_id()
    msg = pack(wire.GET_PROPERTIES, wire.GetPropertiesRequest(properties=["ro.product.device"]))
    service_queue.put(msg)
    resp = wait_response(rid)
    res = {}
    if not resp or not resp.success:
        return res
    for prop in resp.properties:
        res[prop.name] = prop.value
    return res

def on_battery_event(callback):
    fields = [f.name for f in wire.BatteryEvent.DESCRIPTOR.fields]
    def _cb(resp):
        data = dict([(f, getattr(resp, f)) for f in fields])
        callback(data)
    register_eventhook(wire.EVENT_BATTERY, _cb)

def on_rotation_event(callback):
    def _cb(resp):
        callback(resp.rotation)
    register_eventhook(wire.EVENT_ROTATION, _cb)

def on_connectivity_event(callback):
    fields = [f.name for f in wire.ConnectivityEvent.DESCRIPTOR.fields]
    def _cb(resp):
        data = dict([(f, getattr(resp, f)) for f in fields])
        callback(data)
    register_eventhook(wire.EVENT_CONNECTIVITY, _cb)



#-*- encoding: utf-8 -*-

import os
import os.path
import json
import socket
import webbrowser
import tornado.ioloop
import tornado.web
import tornado.websocket
import traceback
import time

from tornado.web import StaticFileHandler

import atx
from atx.record.android import AndroidRecorder

__dir__ = os.path.dirname(os.path.abspath(__file__))

cm = None

class CaseManager(object):
    def __init__(self, basedir):
        self.basedir = basedir

        # record object
        record = AndroidRecorder(None, basedir)
        obj = {}
        with open(os.path.join(basedir, 'frames', 'frames.json')) as f:
            obj = json.load(f)
        record.device_info = obj['device']
        record.frames = obj['frames']
        self.record = record

        # case
        self.casepath = os.path.join(basedir, 'case', 'case.json')
        self.case = []
        with open(self.casepath) as f:
            self.case = json.load(f)

        self._env = None

    def save_case(self, data):
        self.case = json.loads(data)
        with open(self.casepath, 'w') as f:
            json.dump(self.case, f, indent=2)

        # generate code
        try:
            AndroidRecorder.process_casefile(self.basedir)
            return True
        except:
            traceback.print_exc()
            return False

    def build_exec_env(self):
        d = atx.connect()
        d.image_path.append(os.path.join(self.basedir, 'case'))
        self._env = {'time': time, 'd': d}

    def run_step(self, frameidx):
        if self._env is None:
            self.build_exec_env()
        for row in self.case:
            if row['frameidx'] == frameidx:
                code = self.record.process_draft(row)
                print 'running', row, code
                try:
                    cobj = compile(code, '<string>', 'exec')
                    ret = eval(cobj, None, self._env)
                except Exception as e:
                    return str(e)
                else:
                    return 'ok, return value: %s' % str(ret)
        return 'Not Found.'

class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.redirect('/index.html')

class CaseHandler(tornado.web.RequestHandler):

    def get(self, *args):
        self.write(json.dumps(cm.case))
        self.finish()

    def post(self, *args):
        data = self.request.arguments['data'][0] ## get the string
        ok = cm.save_case(data)
        self.write(json.dumps({'success':ok}))

class CaseRunnerHandler(tornado.websocket.WebSocketHandler):

    def check_origin(self, origin):
        return True

    def open(self, *args, **kwargs):
        print 'Websocket connnected.'

    def on_close(self):
        print 'Websocket closed.'

    def on_message(self, message):
        print 'received:', message
        try:
            data = json.loads(message)
        except Exception as e:
            self.write_message({'err': str(e)})
            return

        action = data['action']
        frame = data.get('frame')
        if action in ('run_all', 'run_step'):
            msg = cm.run_step(frame)
            self.write_message({'action':action, 'frame':frame, 'msg': msg})
        else:
            self.write_message({'err': 'unknown action %s' % action})

def get_valid_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    port = s.getsockname()[1]
    s.close()
    return port

def run(basedir, port=8000):
    global cm
    basedir = os.path.abspath(basedir)
    cm = CaseManager(basedir)
    application = tornado.web.Application([
        (r'/', MainHandler),
        (r'/frames/(.*)', StaticFileHandler, {'path':os.path.join(basedir, 'frames')}),
        (r'/case(.*)', CaseHandler),
        (r'/run', CaseRunnerHandler),
        (r'/(.*)', StaticFileHandler, {'path':os.path.join(__dir__, 'site')}),
    ], autoreload=True, static_hash_cache=False)

    if port is None:
        port = get_valid_port()
    webbrowser.open('http://127.0.0.1:%s' % port, new=2)

    application.listen(port)
    print 'Listen on', port
    print 'WorkDir:', basedir
    print 'Press Ctrl+C to stop...'
    try:
        tornado.ioloop.IOLoop.instance().start()
    except:
        print 'Done'

if __name__ == '__main__':
    run('testcase')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import time
import mock

import pytest
import atx
from PIL import Image

d = atx.connect(platform='dummy')

def setup_function(f):
    d.resolution = (1280, 720)


def test_setget_resolution():
    assert d.resolution == (720, 1280)

    d.resolution = None # None is also OK to set
    assert d.resolution is None

    d.resolution = (200, 400)
    assert d.resolution == (200, 400)

    with pytest.raises(TypeError):
        d.resolution = [1, 3]
    with pytest.raises(TypeError):
        d.resolution = 720
    assert d.resolution == (200, 400)


def teardown_function(f):
    print('teardown')


def test_screenshot_normal():
    screen = d.screenshot()
    assert screen is not None


def test_screenshot_first_fail():
    d._fail_first_screenshot = True
    screen = d.screenshot()
    assert screen is not None


def test_screenshot_always_fail():
    import types
    d = atx.connect(platform='dummy')

    def raise_ioerror(self):
        raise IOError('error of io')

    d._take_screenshot = types.MethodType(raise_ioerror, d)
    with pytest.raises(IOError):
        d.screenshot()


def test_hook_screenshot():
    called = [False]

    def hook(event):
        print('event', event)
        called[0] = True

    d.add_listener(hook, atx.EVENT_SCREENSHOT)
    d.screenshot()
    assert called[0] == True

#def test_cloudtest_hook():
#    cloudtest.record_operation(d)
#    d.screenshot()

def test_region_screenshot():
    nd = d.region(atx.Bounds(100, 100, 600, 300))
    rs = nd.region_screenshot()
    assert rs is not None
    assert rs.size == (500, 200)

def test_assert_exists():
    d.assert_exists('media/system-app.png')

    with pytest.raises(atx.AssertExistsError):
        d.assert_exists('media/haima.png', timeout=0.1)

def test_click():
    d.click(50, 70)
    assert d.last_click == (50, 70)

def test_click_image():
    """ require aircv installed """
    d.click_image('media/system-app.png')
    assert d.last_click == (139, 299)

def test_click_image_offset1():
    d.click_image(atx.Pattern('media/system-app.png'))
    assert d.last_click == (139, 299)

def test_click_image_offset2():
    d.click_image(atx.Pattern('media/system-app.png', offset=(0.5, 0.5)))
    w, h = Image.open('media/system-app.png').size
    assert d.last_click == (139+w/2, 299+h/2)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# check symbols are from:
# http://www.i2symbol.com/symbols/check

import os
import subprocess

from colorama import init, Fore, Back, Style
init()


def print_info(message, success):
    fore_color = Fore.GREEN
    symbol = '✔' if os.name != 'nt' else '[GOOD]'
    if not success:
        fore_color = Fore.RED
        symbol = '✘' if os.name != 'nt' else '[FAIL]'

    print(fore_color + "{symbol} {message}".format(symbol=symbol, message=message) + Style.RESET_ALL)   


def main():
    # check if adb exists
    try:
        subprocess.call(['adb', 'version'])
        print_info("adb found in env PATH", True)
    except OSError:
        print_info("adb not found in env PATH", False)


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module is to make mobile test more easily
"""

from __future__ import absolute_import

import os
import sys
import six

import pkg_resources
try:
    version = pkg_resources.get_distribution("atx").version
except pkg_resources.DistributionNotFound:
    version = 'unknown'

from atx.consts import *
from atx.errors import *
from atx.drivers import Pattern, Bounds, ImageCrop


def _connect_url(*args):
    if len(args) == 0:
        return os.getenv('ATX_CONNECT_URL')
    return args[0]


def _detect_platform(connect_url):
    if os.getenv('ATX_PLATFORM'):
        return os.getenv('ATX_PLATFORM')

    if not connect_url: # None or ""
        return 'android'
    elif connect_url.startswith('http://'): # WDA use http url as connect str
        return 'ios'
    else:
        return 'android'


def connect(*args, **kwargs):
    """Connect to a device, and return its object
    Args:
        platform: string one of <android|ios|windows>
        
    Returns:
        None

    Raises:
        SyntaxError, EnvironmentError
    """
    connect_url = _connect_url(*args)
    platform = kwargs.pop('platform', _detect_platform(connect_url))

    cls = None
    if platform == 'android':
        os.environ['JSONRPC_TIMEOUT'] = "60" # default is 90s which is too long.
        devcls = __import__('atx.drivers.android')
        cls = devcls.drivers.android.AndroidDevice
    elif platform == 'windows':
        devcls = __import__('atx.drivers.windows')
        cls = devcls.drivers.windows.WindowsDevice
    elif platform == 'ios':
        devcls = __import__('atx.drivers.ios_webdriveragent')
        cls = devcls.drivers.ios_webdriveragent.IOSDevice
    elif platform == 'webdriver':
        devcls = __import__('atx.drivers.webdriver')
        cls = devcls.drivers.webdriver.WebDriver
    elif platform == 'dummy': # for py.test use
        devcls = __import__('atx.drivers.dummy')
        cls = devcls.drivers.dummy.DummyDevice
    
    if cls is None:
        raise SyntaxError('Platform: %s not exists' % platform)

    c = cls(connect_url, **kwargs)
    c.platform = platform
    return c


# def _sig_handler(signum, frame):
#     print >>sys.stderr, 'Signal INT catched !!!'
#     sys.exit(1)
# signal.signal(signal.SIGINT, _sig_handler)



#-*- encoding: utf-8 -*-
"""
Windows application as a device
"""

from __future__ import absolute_import

import os
import cv2
import ctypes
import numpy as np
import time
import win32con
import win32api
import win32gui
import win32process

from PIL import Image
from ctypes import wintypes, windll
from collections import namedtuple

from atx.drivers import Bounds, Display
from atx.drivers.mixin import DeviceMixin
from atx.errors import WindowsAppNotFoundError

# https://msdn.microsoft.com/en-us/library/windows/desktop/dd183376(v=vs.85).aspx
class BITMAPINFOHEADER(ctypes.Structure):
    _fields_ = [
        ('biSize', wintypes.DWORD),
        ('biWidth', wintypes.LONG),
        ('biHeight', wintypes.LONG),
        ('biPlanes', wintypes.WORD),
        ('biBitCount', wintypes.WORD),
        ('biCompression', wintypes.DWORD),
        ('biSizeImage', wintypes.DWORD),
        ('biXPelsPerMeter', wintypes.LONG),
        ('biYPelsPerMeter', wintypes.LONG),
        ('biClrUsed', wintypes.DWORD),
        ('biClrImportant', wintypes.DWORD)
    ]

def find_process_id(exe_file):
    exe_file = os.path.normpath(exe_file).lower()
    command = "wmic process get processid,commandline"
    for line in os.popen(command).read().lower().splitlines():
        line = line.strip()
        if not line:
            continue
        line = line.split()
        pid = line[-1]
        cmd = " ".join(line[:-1])
        if not cmd:
            continue
        elif cmd.startswith("'") or cmd.startswith('"'):
            pos = cmd.find(cmd[0], 1)
            cmd = cmd[1:pos]
        else:
            cmd = cmd.split()[0]

        if exe_file == cmd:
            return int(pid)

Rect = namedtuple('Rect', ('left', 'top', 'right', 'bottom'))
Size = namedtuple('Size', ('width', 'height'))

class Window(object):
    """A interface of windows' window display zone.

    Args:
        window_name: the text on window border 
        exe_file: the path to windows executable
        exclude_border: count the border in display zone or not. 
            Default is True.

    Attributes:
        screen: a PIL Image object of current display zone.
        rect: (left, top, right, bottom) of the display zone.
        size: (width, height) of the display zone.

    Methods:
        norm_position((x,y)) --> (x0, y0) 
            normalize coordinates relative to window's rect.
    """

    def __init__(self, window_name=None, exe_file=None, exclude_border=True):
        hwnd = 0

        # first check window_name
        if window_name is not None:
            hwnd = win32gui.FindWindow(None, window_name)
            if hwnd == 0:
                def callback(h, extra):
                    if window_name in win32gui.GetWindowText(h):
                        extra.append(h)
                    return True
                extra = []
                win32gui.EnumWindows(callback, extra)
                if extra: hwnd = extra[0]
            if hwnd == 0:
                raise WindowsAppNotFoundError("Windows Application <%s> not found!" % window_name)

        # check exe_file by checking all processes current running.
        elif exe_file is not None:
            pid = find_process_id(exe_file)
            if pid is not None:
                def callback(h, extra):
                    if win32gui.IsWindowVisible(h) and win32gui.IsWindowEnabled(h):
                        _, p = win32process.GetWindowThreadProcessId(h)
                        if p == pid:
                            extra.append(h)
                        return True
                    return True
                extra = []
                win32gui.EnumWindows(callback, extra)
                #TODO: get main window from all windows.
                if extra: hwnd = extra[0]
            if hwnd == 0:
                raise WindowsAppNotFoundError("Windows Application <%s> is not running!" % exe_file)

        # if window_name & exe_file both are None, use the screen.
        if hwnd == 0:
            hwnd = win32gui.GetDesktopWindow()

        self.hwnd = hwnd
        self.exclude_border = exclude_border

    @property
    def rect(self):
        hwnd = self.hwnd
        if not self.exclude_border:
            left, top, right, bottom = win32gui.GetWindowRect(hwnd)
        else:
            _left, _top, _right, _bottom = win32gui.GetClientRect(hwnd)
            left, top = win32gui.ClientToScreen(hwnd, (_left, _top))
            right, bottom = win32gui.ClientToScreen(hwnd, (_right, _bottom))
        return Rect(left, top, right, bottom)

    @property
    def size(self):
        left, top, right, bottom = self.rect
        return Size(right-left, bottom-top)

    def norm_position(self, pos):
        left, top, right, bottom = self.rect
        _left, _top = pos
        x, y = _left-left, _top-top
        if _left > right:
            x = -1
        if _top > bottom:
            y = -1
        return (x, y)

    @property
    def screen(self):
        """PIL Image of current window screen. (the window must be on the top)
        reference: https://msdn.microsoft.com/en-us/library/dd183402(v=vs.85).aspx"""
        # opengl windows cannot get from it's hwnd, so we use the screen
        hwnd = win32gui.GetDesktopWindow()

        # get window size and offset
        left, top, right, bottom = self.rect
        width, height = right-left, bottom-top

        # the device context of the window 
        hdcwin = win32gui.GetWindowDC(hwnd)
        # make a temporary dc
        hdcmem = win32gui.CreateCompatibleDC(hdcwin)
        # make a temporary bitmap in memory, this is a PyHANDLE object
        hbmp = win32gui.CreateCompatibleBitmap(hdcwin, width, height)
        # select bitmap for temporary dc
        win32gui.SelectObject(hdcmem, hbmp)
        # copy bits to temporary dc
        win32gui.BitBlt(hdcmem, 0, 0, width, height, 
                        hdcwin, left, top, win32con.SRCCOPY)
        # check the bitmap object infomation
        bmp = win32gui.GetObject(hbmp)

        bi = BITMAPINFOHEADER()
        bi.biSize = ctypes.sizeof(BITMAPINFOHEADER)
        bi.biWidth = bmp.bmWidth
        bi.biHeight = bmp.bmHeight
        bi.biPlanes = bmp.bmPlanes
        bi.biBitCount = bmp.bmBitsPixel
        bi.biCompression = 0 # BI_RGB
        bi.biSizeImage = 0
        bi.biXPelsPerMeter = 0
        bi.biYPelsPerMeter = 0
        bi.biClrUsed = 0
        bi.biClrImportant = 0

        # calculate total size for bits
        pixel = bmp.bmBitsPixel
        size = ((bmp.bmWidth * pixel + pixel - 1)/pixel) * 4 * bmp.bmHeight
        buf = (ctypes.c_char * size)()

        # read bits into buffer
        windll.gdi32.GetDIBits(hdcmem, hbmp.handle, 0, bmp.bmHeight, buf, ctypes.byref(bi), win32con.DIB_RGB_COLORS)

        # make a PIL Image
        img = Image.frombuffer('RGB', (bmp.bmWidth, bmp.bmHeight), buf, 'raw', 'BGRX', 0, 1)
        img = img.transpose(Image.FLIP_TOP_BOTTOM)

        # cleanup
        win32gui.DeleteObject(hbmp)
        win32gui.DeleteObject(hdcmem)
        win32gui.ReleaseDC(hwnd, hdcwin)

        return img

    def _screenshot(self, filepath):
        dirpath = os.path.dirname(os.path.abspath(filepath))
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        self.screen.save(filepath)

    @property
    def screen_cv2(self):
        """cv2 Image of current window screen"""
        pil_image = self.screen.convert('RGB')
        cv2_image = np.array(pil_image)
        pil_image.close()
        # Convert RGB to BGR 
        cv2_image = cv2_image[:, :, ::-1]
        return cv2_image

    def _screenshot_cv2(self, filepath):
        dirpath = os.path.dirname(os.path.abspath(filepath))
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        cv2.imwrite(filepath, self.screen_cv2)

    def set_foreground(self):
        win32gui.SetForegroundWindow(self.hwnd)

    def _input_left_mouse(self, x, y):
        left, top, right, bottom = self.rect
        width, height = right - left, bottom - top
        if x < 0 or x > width or y < 0 or y > height:
            return

        win32gui.SetForegroundWindow(self.hwnd)
        pos = win32gui.GetCursorPos()
        win32api.SetCursorPos((left+x, top+y))
        win32api.mouse_event(win32con.MOUSEEVENTF_LEFTDOWN, 0, 0)
        win32api.Sleep(100) #ms
        win32api.mouse_event(win32con.MOUSEEVENTF_LEFTUP, 0, 0)
        win32api.Sleep(100) #ms
        # win32api.SetCursorPos(pos)


class FrozenWindow(Window):
    """Non-resizable Window, use lots of cached properties"""

    def __init__(self, *args, **kwargs):
        Window.__init__(self, *args, **kwargs)
        self.__init_rect_size()
        self.__init_screen_handles()

    def __init_rect_size(self):
        hwnd = self.hwnd
        left, top, right, bottom = win32gui.GetWindowRect(hwnd)
        if self.exclude_border:
            _left, _top, _right, _bottom = win32gui.GetClientRect(hwnd)
            left, top = win32gui.ClientToScreen(hwnd, (_left, _top))
            right, bottom = win32gui.ClientToScreen(hwnd, (_right, _bottom))
        self._rect = Rect(left, top, right, bottom)
        self._size = Size(right-left, bottom-top)

    @property 
    def rect(self):
        return self._rect

    @property
    def size(self):
        return self._size

    def __init_screen_handles(self):
        # opengl windows cannot get from it's hwnd, so we use the screen
        hwnd = win32gui.GetDesktopWindow()
        # get screen size and offset
        left, top, right, bottom = self.rect
        width, height = right-left, bottom-top
        # the device context of the window 
        hdcwin = win32gui.GetWindowDC(hwnd)
        # make a temporary dc
        hdcmem = win32gui.CreateCompatibleDC(hdcwin)
        # make a temporary bitmap in memory, this is a PyHANDLE object
        hbmp = win32gui.CreateCompatibleBitmap(hdcwin, width, height)
        # select bitmap for temporary dc
        win32gui.SelectObject(hdcmem, hbmp)
        # check the bitmap object infomation
        bmp = win32gui.GetObject(hbmp)
        bi = BITMAPINFOHEADER()
        bi.biSize = ctypes.sizeof(BITMAPINFOHEADER)
        bi.biWidth = bmp.bmWidth
        bi.biHeight = bmp.bmHeight
        bi.biPlanes = bmp.bmPlanes
        bi.biBitCount = bmp.bmBitsPixel
        bi.biCompression = 0 # BI_RGB
        bi.biSizeImage = 0
        bi.biXPelsPerMeter = 0
        bi.biYPelsPerMeter = 0
        bi.biClrUsed = 0
        bi.biClrImportant = 0
        # calculate total size for bits
        pixel = bmp.bmBitsPixel
        size = ((bmp.bmWidth * pixel + pixel - 1)/pixel) * 4 * bmp.bmHeight
        buf = (ctypes.c_char * size)()

        self._hdcwin = hdcwin
        self._hdcmem = hdcmem
        self._bi = bi
        self._hbmp = hbmp
        self._buf = buf

    @property
    def screen(self):
        """PIL Image of current window screen.
        reference: https://msdn.microsoft.com/en-us/library/dd183402(v=vs.85).aspx"""
        hwnd = win32gui.GetDesktopWindow()
        left, top, right, bottom = self.rect
        width, height = right-left, bottom-top
        # copy bits to temporary dc
        win32gui.BitBlt(self._hdcmem, 0, 0, width, height, 
                        self._hdcwin, left, top, win32con.SRCCOPY)
        # read bits into buffer
        windll.gdi32.GetDIBits(self._hdcmem, self._hbmp.handle, 0, height, self._buf, ctypes.byref(self._bi), win32con.DIB_RGB_COLORS)
        # make a PIL Image
        img = Image.frombuffer('RGB', (width, height), self._buf, 'raw', 'BGRX', 0, 1)
        img = img.transpose(Image.FLIP_TOP_BOTTOM)
        return img

    @property
    def screen_cv2(self):
        """cv2 Image of current window screen"""
        hwnd = win32gui.GetDesktopWindow()
        left, top, right, bottom = self.rect
        width, height = right-left, bottom-top
        # copy bits to temporary dc
        win32gui.BitBlt(self._hdcmem, 0, 0, width, height, 
                        self._hdcwin, left, top, win32con.SRCCOPY)
        # read bits into buffer
        windll.gdi32.GetDIBits(self._hdcmem, self._hbmp.handle, 0, height, self._buf, ctypes.byref(self._bi), win32con.DIB_RGB_COLORS)
        # make a cv2 Image
        arr = np.fromstring(self._buf, dtype=np.uint8)
        img = arr.reshape(height, width, 4)
        img = img[::-1,:, 0:3]
        return img

    def __del__(self):
        # cleanup
        hwnd = win32gui.GetDesktopWindow()
        win32gui.DeleteObject(self._hbmp)
        win32gui.DeleteObject(self._hdcmem)
        win32gui.ReleaseDC(hwnd, self._hdcwin)

class WindowsDevice(DeviceMixin):
    def __init__(self, winclass=FrozenWindow, **kwargs):
        DeviceMixin.__init__(self)
        self._win = winclass(**kwargs)

    def __getattr__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except:
            return getattr(self._win, attr)

    def screenshot(self, filename=None):
        """Take screen snapshot

        Args:
            filename: filename where save to, optional

        Returns:
            PIL.Image object

        Raises:
            TypeError, IOError
        """
        if filename:
            self._screenshot(filename)
        return self.screen

    def screenshot_cv2(self, filename=None):
        """Take screen snapshot

        Args:
            filename: filename where save to, optional

        Returns:
            numpy.ndarray object as return by cv2.imread

        Raises:
            IOError
        """
        if filename:
            self._screenshot_cv2(filename)
        return self.screen_cv2

    def click(self, x, y):
        """Simulate click within window screen.

        Args:
            x, y: int, pixel distance from window (left, top) as origin

        Returns:
            None
        """
        print 'click at', x, y
        self._input_left_mouse(x, y)

    def text(self, text):
        """Simulate text input to window.

        Args:
            text: string

        Returns:
            None
        """

    @property
    def display(self):
        """Display size in pixels."""
        w, h = self.size
        return Display(w, h)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import re
import socket
try:
    import subprocess32 as subprocess
except:
    import subprocess

from atx.adbkit.device import Device


LOCAL_PORT = 10300
_init_local_port = LOCAL_PORT - 1

def next_local_port(adb_host=None):
    """ find avaliable free port """
    def is_port_listening(port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((str(adb_host) if adb_host else '127.0.0.1', port))
        s.close()
        return result == 0
    global _init_local_port
    _init_local_port = _init_local_port + 1 if _init_local_port < 32764 else LOCAL_PORT
    while is_port_listening(_init_local_port):
        _init_local_port += 1
    return _init_local_port


class Client(object):
    __adb_cmd = None

    def __init__(self, host='127.0.0.1', port=5037):
        """
        Args:
            - host: adb server host, default 127.0.0.1
            - port: adb server port, default 5037
        """
        self._host = host or '127.0.0.1'
        self._port = port or 5037

    @property
    def server_host(self):
        return self._host
    
    @classmethod
    def adb_path(cls):
        """return adb binary full path"""
        if cls.__adb_cmd is None:
            if "ANDROID_HOME" in os.environ:
                filename = "adb.exe" if os.name == 'nt' else "adb"
                adb_dir = os.path.join(os.environ["ANDROID_HOME"], "platform-tools")
                adb_cmd = os.path.join(adb_dir, filename)
                if not os.path.exists(adb_cmd):
                    raise EnvironmentError(
                        "Adb not found in $ANDROID_HOME/platform-tools path: %s." % adb_dir)
            else:
                import distutils
                if "spawn" not in dir(distutils):
                    import distutils.spawn
                adb_cmd = distutils.spawn.find_executable("adb")
                if adb_cmd:
                    adb_cmd = os.path.realpath(adb_cmd)
                else:
                    raise EnvironmentError("$ANDROID_HOME environment not set.")
            cls.__adb_cmd = adb_cmd
        return cls.__adb_cmd

    @property    
    def _host_port_args(self):
        args = []
        if self._host and self._host != '127.0.0.1':
            args += ['-H', self._host]
        if self._port:
            args += ['-P', str(self._port)]
        return args

    def raw_cmd(self, *args, **kwargs):
        '''adb command. return the subprocess.Popen object.'''
        cmds = [self.adb_path()] + self._host_port_args + list(args)
        kwargs['stdout'] = kwargs.get('stdout', subprocess.PIPE)
        kwargs['stderr'] = kwargs.get('stderr', subprocess.PIPE)
        # if os.name != "nt":
        #     cmd_line = [" ".join(cmd_line)]
        return subprocess.Popen(cmds, **kwargs)

    def run_cmd(self, *args, **kwargs):
        p = self.raw_cmd(*args, **kwargs)
        return p.communicate()[0].decode('utf-8').replace('\r\n', '\n')

    def devices(self):
        '''get a dict of attached devices. key is the device serial, value is device name.'''
        out = self.run_cmd('devices') #subprocess.check_output([self.adb_path(), 'devices']).decode("utf-8")
        if 'adb server is out of date' in out:
            out = self.run_cmd('devices')
        match = "List of devices attached"
        index = out.find(match)
        if index < 0:
            raise EnvironmentError("adb is not working.")
        return dict([s.split("\t") for s in out[index + len(match):].strip().splitlines() 
                if s.strip() and not s.strip().startswith('*')])

    def version(self):
        '''
        Get version of current adb
        Return example:
            [u'1.0.32', u'1', u'0', u'32']
        '''
        '''adb version'''
        match = re.search(r"(\d+)\.(\d+)\.(\d+)", self.run_cmd("version"))
        return [match.group(i) for i in range(4)]

    def device(self, serial=None):
        devices = self.devices()
        if not serial:
            if devices:
                if len(devices) is 1:
                    serial = list(devices.keys())[0]
                else:
                    raise EnvironmentError("Multiple devices attached but default android serial not set.")
            else:
                raise EnvironmentError("Device not attached.")
        else:
            if serial not in devices:
                raise EnvironmentError("Device(%s) not attached." % serial)

        if devices[serial] != 'device':
            raise EnvironmentError("Device(%s) is not ready. status(%s)." % 
                (serial, devices[serial]))
        return Device(self, serial)

    def connect(self, addr):
        '''
        Call adb connect
        Return true when connect success
        '''
        if addr.find(':') == -1:
            addr += ':5555'
        output = self.run_cmd('connect', addr)
        return 'unable to connect' not in output

    def disconnect(self, addr):
        ''' disconnect device '''
        return self.raw_cmd('disconnect', addr).wait()

    def forward_list(self):
        '''
        adb forward --list
        TODO: not tested
        '''
        version = self.version()
        if int(version[1]) <= 1 and int(version[2]) <= 0 and int(version[3]) < 31:
            raise EnvironmentError("Low adb version.")
        lines = self.run_cmd("forward", "--list").strip().splitlines()
        return [line.strip().split() for line in lines]

    def forward(self, serial, local_port, remote_port=None):
        '''
        adb port forward. return local_port
        TODO: not tested
        '''
        # Shift args, because remote_port is required while local_port is optional
        if remote_port is None:
            remote_port, local_port = local_port, None
        if not local_port:
            for s, lp, rp in self.forward_list():
                if s == serial and rp == 'tcp:%d' % remote_port:
                    return int(lp[4:])
            return self.forward(serial, next_local_port(self.server_host), remote_port)
        else:
            print(serial, remote_port, local_port)
            self.raw_cmd("-s", serial, "forward", "tcp:%d" % local_port, "tcp:%d" % remote_port).wait()
            return local_port


if __name__ == '__main__':
    adb = Client()
    print(adb.devices())
    print(adb.version())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import time


code = ''
lineno = 0

def interpreter(*args):
    print 'I:', args
    time.sleep(0.5)

for line in open('simple-tcp-proxy.py'):
    lineno += 1
    contain_code = False
    if line.strip() and not line.strip().startswith('#'):
        contain_code = True
    if contain_code:
        code += 'interpreter({}, {})\n'.format(lineno, json.dumps(line.strip()))
    code += line

print code
exec(code, {'interpreter': interpreter})

# print line.rstrip()



# coding: utf-8

import os
import time
import atx
import logging
from atx import consts
print 'Version:', atx.version

d = atx.connect(None)

# 你好

def screenshot():
    start = time.time()
    #d.screenshot_method = consts.SCREENSHOT_METHOD_MINICAP
    d.screenshot('ttt.png')
    print time.time() - start

def start_app():
    d.start_app('com.netease.txx')

def stop_app():
    d.stop_app('com.netease.txx', clear=True)

def touch():
    d.screenshot_method = consts.SCREENSHOT_METHOD_MINICAP
    d.touch_image('button.png')

if __name__ == '__main__':
    log = logging.getLogger('atx')
    log.setLevel(logging.DEBUG)

    # d.screen.off()

    #def foo(evt):
        #print 'good', evt
        #d.click(*evt.pos)

    #with d.watch('simulator', 10) as w:
        #w.on(atx.Pattern("mmm.png", offset=(-79, -13))).do(foo).quit()
    # # stop_app()
    #print 'inside'
    #screenshot()
    # print d.dump_nodes()
    # w.on('setting.png', atx.Watcher.ACTION_TOUCH)
    # w.on('common.png', atx.Watcher.ACTION_TOUCH)

    # wid = d.add_watcher(w)
    # d.del_watcher(wid)
    # while 1:
    #     screenshot()
    # screenshot()
    # touch()



#-*- encoding: utf-8 -*-

import os
import sys
import time

from atx.drivers.windows import WindowsDevice, find_process_id
from atx.drivers.android import AndroidDevice
from atx.cmds.record import RecorderGUI

def get_calc_win():
    exe_file = "C:\\Windows\\System32\\calc.exe"
    if not find_process_id(exe_file):
        os.startfile(exe_file)
        time.sleep(3)

    win = WindowsDevice(exe_file=exe_file)
    print "window handle", hex(win.hwnd)
    return win

def get_game_win():
    window_name = "MyLuaGame"
    win = WindowsDevice(window_name=window_name)
    print "window handle", hex(win.hwnd)
    return win

def get_android_dev():
    dev = AndroidDevice()
    print 'android devcie', dev._serial
    return dev

if len(sys.argv) > 1 and sys.argv[1] == 'win':
    get_device = get_calc_win
else:
    get_device = get_android_dev

def main():
    dev = get_device()
    print "display size %dx%d" % dev.display

    r = RecorderGUI(dev)
    r.mainloop()

if __name__ == '__main__':
    main()


#! /usr/bin/env python
# -*- coding: utf-8 -*-

import atexit
import os
import re
import shutil
import sys
import time
import tempfile

import tqdm

from atx import logutils
from atx import adbkit
from atx.cmds import utils

try:
    import subprocess32 as subprocess
except:
    import subprocess


log = logutils.getLogger('install')
DEFAULT_REMOTE_PATH = '/data/local/tmp/_atx_tmp.apk'
__apks = {
    'utf7ime': 'http://o8oookdsx.qnssl.com/android_unicode_ime-debug.apk',
    'atx-assistant': 'https://o8oookdsx.qnssl.com/atx-assistant-1.0.4.apk',
}


def clean(tmpdir):
    log.info('Remove temp directory')
    shutil.rmtree(tmpdir)


def adb_pushfile(adb, filepath, remote_path):
    filesize = os.path.getsize(filepath)
    pb = tqdm.tqdm(unit='B', unit_scale=True, total=filesize)
    p = adb.raw_cmd('push', filepath, remote_path)

    while True:
        try:
            p.wait(0.5)
        except subprocess.TimeoutExpired:
            pb.n = get_file_size(adb, remote_path)
            pb.refresh()
            # log.info("Progress %dM/%dM", get_file_size(remote_path) >>20, filesize >>20)
            pass
        except (KeyboardInterrupt, SystemExit):
            p.kill()
            raise
        except:
            raise
        else:
            # log.info("Success pushed into device")
            break
    pb.close()


def get_file_size(adb, remote_path):
    try:
        output = adb.run_cmd('shell', 'ls', '-l', remote_path)
        m = re.search(r'\s(\d+)', output)
        if not m:
            return 0
        return int(m.group(1))
    except subprocess.CalledProcessError as e:
        log.warn("call error: %s", e)
        time.sleep(.1)
        return 0


def adb_remove(adb, path):
    p = adb.raw_cmd('shell', 'rm', path)
    stdout, stderr = p.communicate()
    if stdout or stderr:
        log.warn('%s\n%s', stdout, stderr)


def adb_install(adb, remote_path):
    stdout = adb.run_cmd('shell', 'pm', 'install', '-rt', remote_path)
    # stdout, _ = p.communicate()
    if stdout.find('Success') == -1:
        raise IOError("Adb install failed: %s" % stdout)


def adb_must_install(adb, remote_path, package_name):
    try:
        adb_install(adb, remote_path)
    except IOError:
        log.info("Remove already installed app: %s", package_name)
        adb.raw_cmd('uninstall', package_name).wait()
        adb_install(adb, remote_path)


def main(path, serial=None, host=None, port=None, start=False):
    adb = adbkit.Client(host, port).device(serial) #adbutils.Adb(serial, host, port)
    
    # use qiniu paths
    if __apks.get(path):
        path = __apks.get(path)

    if re.match(r'^https?://', path):
        tmpdir = tempfile.mkdtemp(prefix='atx-install-')
        log.info("Create temp directory: %s", tmpdir)
        
        # FIXME(ssx): will not called when Ctrl+C pressed in windows git-bash
        atexit.register(clean, tmpdir)

        urlpath = path
        target = os.path.join(tmpdir, '_tmp.apk')
        path = target
        log.info("Download from: %s", urlpath)
        utils.http_download(urlpath, target)

    import atx.androaxml as apkparse
    package_name, main_activity = apkparse.parse_apk(path)
    log.info("APK package name: %s", package_name)
    log.info("APK main activity: %s", main_activity)

    log.info("Push file to android device")
    adb_pushfile(adb, path, DEFAULT_REMOTE_PATH)

    log.info("Install ..., will take a few seconds")
    adb_must_install(adb, DEFAULT_REMOTE_PATH, package_name)
    log.info("Remove _tmp.apk")
    adb_remove(adb, DEFAULT_REMOTE_PATH)

    if start:
        log.info("Start app '%s'" % package_name)
        adb.raw_cmd('shell', 'am', 'start', '-n', package_name+'/'+main_activity).wait()
    log.info("Success")


# -*- coding: utf-8 -*-
# hook only ui related events via uiautomator.
# the basic idea is find gesture target via postion.
# so we should simulate the layout via dumped ui tree.
# The ui hierarchy needs to be dumped repeatedly
# and the process costs a rather long time.
# after each gesture we need to refresh the tree.


import re
import cv2
import collections
import xml.dom.minidom
import numpy as np

from atx.drivers import Bounds
__UINode = collections.namedtuple('UINode', [
    'xml', 'children', 'depth',
    'index', 'text', 'resourceId',
    'className', 'packageName', 'description',
    'checkable', 'checked', 'clickable', 'enabled', 'focusable', 'focused',
    'scrollable', 'longClickable', 'password', 'selected',
    'bounds'])
# make it hashable & traceble
class UINode(__UINode):
    parent=None
    iterindex = None
    def __hash__(self):
        return id(self)

# ignore text/description contains punctuation
txt_pat = re.compile(ur'^[a-zA-Z0-9 \u4e00-\u9fa5]+$')

def parse_bounds(text):
    m = re.match(r'\[(\d+),(\d+)\]\[(\d+),(\d+)\]', text)
    if m is None:
        return None
    return Bounds(*map(int, m.groups()))

def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")

def convstr(v):
    return v

class AndroidLayout(object):
    def __init__(self):
        self.tree = None
        self.nodes = []
        self.rotation = 0

    def find_node(self, x, y):
        for n in self.nodes:
            if n.enabled and n.bounds.is_inside(x, y):
                return n

    def find_clickable_node(self, x, y):
        for n in self.nodes:
            if n.enabled and n.clickable and n.bounds.is_inside(x, y):
                return n

    def find_scrollable_node(self, x, y):
        for n in self.nodes:
            if n.enabled and n.scrollable and n.bounds.is_inside(x, y):
                return n

    def display(self, showall=False):
        if not self.tree or not self.nodes:
            return
        b = self.tree.bounds
        l, t = b.left, b.top
        w, h = b.right - b.left, b.bottom - b.top
        img = np.zeros((h, w, 3), np.uint8)

        i = 0
        for n in self.nodes[::-1]:
            if not showall and not n.clickable and not n.scrollable: continue
            b = n.bounds
            cv2.rectangle(img, (b.left-l, b.top-t), (b.right-l, b.bottom-t), (83, min(255, i*5), 18), 2)
            i += 1
        return img

    def _parse_xml_node(self, node, depth=0):
        __alias = {
            'class': 'className',
            'package': 'packageName',
            'resource-id': 'resourceId',
            'content-desc': 'description',
            'long-clickable': 'longClickable',
        }

        parsers = {
            'index': int,
            'text': convstr,
            'resourceId': convstr,
            'className': convstr,
            'packageName': convstr,
            'description': convstr,
            'bounds': parse_bounds,
            'checkable': str2bool,
            'checked': str2bool,
            'clickable': str2bool,
            'enabled': str2bool,
            'focusable': str2bool,
            'focused': str2bool,
            'scrollable': str2bool,
            'longClickable': str2bool,
            'password': str2bool,
            'selected': str2bool,
        }
        ks = {}
        for key, value in node.attributes.items():
            key = __alias.get(key, key)
            f = parsers.get(key)
            if value is None:
                ks[key] = None
            elif f:
                ks[key] = f(value)
        for key in parsers.keys():
            ks[key] = ks.get(key)
        ks['children'] = []
        ks['depth'] = depth
        ks['xml'] = node

        return UINode(**ks)

    def parse_xmldata(self, xmldata):
        dom = xml.dom.minidom.parseString(xmldata)
        root = dom.documentElement
        self.rotation = int(root.getAttribute('rotation'))

        idx = [0]
        def walk(node, ui_nodes, depth=0):
            while len(node.childNodes) == 1 and node.getAttribute('bounds') == '':
                node = node.childNodes[0]
                depth += 1
            uinode = self._parse_xml_node(node, depth)
            uinode.iterindex = idx[0]
            idx[0] += 1
            for n in node.childNodes:
                sub = walk(n, ui_nodes, depth+1)
                if sub is not None:
                    uinode.children.append(sub)
                    sub.parent = uinode
            ui_nodes.append(uinode)
            return uinode

        self.nodes = []
        self.tree = walk(root, self.nodes)
        self.nodes.sort(key=lambda x: x.bounds.area)
        self.node_index = [n.iterindex for n in self.nodes]

    def find_selector(self, node):
        '''find condition for locate a node, return (postion_node, condition, order)'''
        return self.__find_selector_by_structure(node)
        # return self.__find_selector_by_attrbutes(node)
        # return self.__find_selector_by_score(node)

    def _filter_nodes(self, cond, nodes=None):
        if nodes is None:
            nodes = self.nodes
        res = []
        for n in nodes:
            match = True
            for k, v in cond.iteritems():
                attr = getattr(n, k)
                if isinstance(v, re._pattern_type) and \
                    isinstance(attr, basestring) and v.match(attr) is None:
                    match = False
                    break
                elif attr != v:
                    match = False
                    break
            if match:
                res.append(n)
        return res

    def get_index_node(self, idx):
        '''get node with iterindex `idx`'''
        idx = self.node_index.index(idx)
        return self.nodes[idx]

    def get_node_selector(self, n, sub=False):
        d = {}
        nodes = self.nodes

        # try text, text in subnode cannot be all numbers
        if n.text and txt_pat.match(n.text) and not (sub and n.text.isdigit()):
            d['text'] = n.text
            nodes = self._filter_nodes(d, nodes)
            if len(nodes) == 1:
                return d, None

        # try className, which is never blank
        d['className'] = n.className
        nodes = self._filter_nodes(d, nodes)
        if len(nodes) == 1:
            return d, None

        # try description, like text
        if n.description and txt_pat.match(n.description) and not (sub and n.description.isdigit()):
            d['description'] = n.description
            nodes = self._filter_nodes(d, nodes)
            if len(nodes) == 1:
                return d, None

        ## try resourceId, but on some phone the resourceId may not be available
        if n.resourceId:
            d['resourceId'] = n.resourceId
            nodes = self._filter_nodes(d, nodes)
            if len(nodes) == 1:
                return d, None

        # return order in selected nodes under condition d.
        return d, nodes.index(n)

    def __find_selector_by_structure(self, node):
        # try itself
        d, order = self.get_node_selector(node)
        if order is None:
            return node, d, None

        # try its non-clickable decendants
        decendants = []
        def walk(n):
            for c in n.children:
                if c.clickable: continue
                decendants.append(c)
                walk(c)
        walk(node)

        choices = []
        for n in decendants:
            sd, sorder = self.get_node_selector(n, True)
            choices.append((sorder or 0, -n.bounds.area, sorder, sd, n)) # add area to sort
        choices.sort()

        if choices:
            return choices[0][-1], choices[0][-2], choices[0][-3]

        # TODO
        # # try if its non-clickable parent
        # def is_decendant(n1, n2):
        #     '''check if n1 is decandant of n2'''
        #     for c in n2.children:
        #         if c == n1 or is_decendant(n1, c):
        #             return True
        #     return False
        #
        # p = node.parent
        # while p and not p.clickable:
        #     pd, porder = self.get_node_selector(p)
        #     if porder is None:
        #         print 'parent node', p.className
        #         for i in range(len(p.children)):
        #             c = p.children[i]
        #             if c == node or is_decendant(node, c):
        #                 print 'child', i, c==node, c.className, c.resourceId, c.text
        #         return pd, porder
        #     p = p.parent

        return node, d, order

    def __find_selector_by_attrbutes(self, node):
        # avoid repeat over same attr

        def attrs(n, name):
            '''get node attribute values'''
            res = set()
            v = getattr(n, name)
            if v: res.add(v)
            for subn in n.children:
                if subn.clickable: continue
                res.update(attrs(subn, name))
            return res

        def is_decendant(n1, n2):
            '''check if n1 is decandant of n2'''
            for c in n2.children:
                if c == n1 or is_decendant(n1, c):
                    return True
            return False

        candidates = {}
        def try_attr(top, attr, ignore_filter=None):
            for value in attrs(top, attr):
                if ignore_filter and ignore_filter(value):
                    continue
                tmp = self._filter_nodes({attr:value})
                if len(tmp) == 1:
                    return tmp[0], {attr:value}
                # save candidates
                for n in tmp:
                    if n == top or is_decendant(n, top):
                        candidates.setdefault(n, {})[attr] = len(tmp)
            return None, None

        # try className
        n, cond = try_attr(node, 'className')
        if n:
            return n, cond, None

        # try anything with a resourceId
        n, cond = try_attr(node, 'resourceId')
        if n:
            return n, cond, None

        # try anything with a text
        n, cond = try_attr(node, 'text', lambda s: txt_pat.match(s) is None)
        if n:
            return n, cond, None

        # try anything with a description
        n, cond = try_attr(node, 'description', lambda s: txt_pat.match(s) is None)
        if n:
            return n, cond, None

        print 'candidates:', candidates.values()

        # try combinations

        return node, self.get_node_selector(node)

    def __find_selector_by_score(self, node):
        # TODO
        # find candidate selectors and give a score
        candidates = {}

        def walk(n):
            info = {'depth': n.depth-node.depth}
            d, o = self.get_node_selector(n)
            info['selector'] = d
            info['order'] = o
            info['score'] = 0
            candidates[n] = info
            for c in n.children:
                if c.clickable: continue
                walk(c)
        walk(node)

        # get top score selector

if __name__ == '__main__':
    # import subprocess
    # subprocess.check_call('adb shell uiautomator dump /data/local/tmp/window_dump.xml')
    # subprocess.check_call('adb pull /data/local/tmp/window_dump.xml')
    # xmldata = open('window_dump.xml').read()

    import time
    import traceback
    import locale

    encoding = locale.getpreferredencoding()

    from uiautomator import device
    device.dump()

    layout = AndroidLayout()
    layout.highlight = np.zeros((1, 1, 3), np.uint8)

    cv2.namedWindow("layout")

    def on_mouse(event, x, y, flags, param):
        layout, downpos, ismove = param

        # record downpos
        if event == cv2.EVENT_LBUTTONDOWN:
            print 'click at', x*2, y*2 # picture is half-sized.
            param[1] = (x, y)
            param[2] = False
            return
        # check if is moving
        if event == cv2.EVENT_MOUSEMOVE:
            if ismove: return
            if downpos is None:
                param[2] = False
                return
            _x, _y = downpos
            if (_x-x)**2 + (_y-y)**2 > 64:
                param[2] = True
            return
        if event != cv2.EVENT_LBUTTONUP:
            return

        # update layout.highlight
        b = layout.tree.bounds
        l, t = b.left, b.top
        w, h = b.right - b.left, b.bottom - b.top
        highlight = np.zeros((h, w, 3), np.uint8)

        if downpos and ismove: # drag
            node = layout.find_scrollable_node(x*2+l, y*2+t)
            print 'scroll to', x*2, y*2
            if node:
                b = node.bounds
                print 'scrollable node', b, node.index, node.className,
                print 'resource_id:', node.resourceId,
                print 'text:', node.text.encode(encoding, 'ignore'),
                print 'desc:', node.description.encode(encoding, 'ignore')
                cv2.rectangle(highlight, (b.left-l, b.top-t), (b.right-l, b.bottom-t), (0,255,255), 4)
        else:
            node = layout.find_clickable_node(x*2+l, y*2+t)
            if node:
                b = node.bounds
                print 'clickable node', b, node.index, node.className,
                print 'resource_id:', node.resourceId,
                print 'text:', node.text.encode(encoding, 'ignore'),
                print 'desc:', node.description.encode(encoding, 'ignore')
                print device(className=node.className, index=node.index).info
                cv2.rectangle(highlight, (b.left-l, b.top-t), (b.right-l, b.bottom-t), (0,0,255), 4)
                cond, order = layout.find_selector(node)
                if cond:
                    print 'selector', cond, order
                    subnode = layout._filter_nodes(cond)[order or 0]
                    b = subnode.bounds
                    cv2.rectangle(highlight, (b.left-l, b.top-t), (b.right-l, b.bottom-t), (0,180,255), 4)

        param[0].highlight = highlight
        param[1], param[2] = None, False

    cv2.setMouseCallback('layout', on_mouse, [layout, None, False])

    tic = time.time()
    count = 0
    package = None
    try:
        while True:
            xmldata = device.dump(pretty=False).encode('utf-8')
            layout.parse_xmldata(xmldata)
            if layout.tree.packageName != package:
                package = layout.tree.packageName
                print "change to", package

            img = layout.display()
            if img.shape == layout.highlight.shape:
                img += layout.highlight

            h, w = img.shape[:2]
            img = cv2.resize(img, (w/2, h/2))
            cv2.imshow('layout', img)
            # key = cv2.waitKey(10)
            # if key == 115:
            #     with open('%d-%s.xml' % (count, package), 'w') as f:
            #         print 'saved', count, package
            #         f.write(xmldata)
            cv2.waitKey(1)
            count += 1
    except:
        traceback.print_exc()

    toc = time.time()
    t = toc - tic
    if count > 0:
        print 'get %d dumps in %f seconds (%f each)' % (count, t, t/count)
    else:
        print 'get nothing.'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'hzsunshx'

#import numpy as np
import cv2
#from matplotlib import pyplot as plt


MIN_MATCH_COUNT = 8

def _middlePoint(pts):
    def add(p1, p2):
        return (p1[0]+p2[0], p1[1]+p2[1])
    def distance(p1, p2):
        import math
        l2 = (p1[0]-p2[0])*(p1[0]-p2[0]) + (p1[1]-p2[1])*(p1[1]-p2[1])
        return math.sqrt(l2)
    # debug
    for p in pts:
        print 'Point:', p.pt
    length = len(pts)
    sumx, sumy = reduce(add, [p.pt for p in pts])
    point = sumx/length, sumy/length
    print 'step1: result=', point

    # filter out ok points
    avg_distance = sum([distance(point, p.pt) for p in pts])/length
    print 'avg distance=', avg_distance
    good = []
    sumx, sumy = 0.0, 0.0
    for p in pts:
        print 'point: %s, distance: %.2f' %(p.pt, distance(p.pt, point))
        if distance(p.pt, point) < 1.2*avg_distance:
            good.append(p.pt)
            sumx += p.pt[0]
            sumy += p.pt[1]
        else:
            print 'not good', p.pt
    print 'step1: result=', point
    point = map(long, (sumx/len(good), sumy/len(good)))
    print 'step2: point=', point
    return point

def find_image_position(origin='origin.png', query='query.png', outfile=None):
    '''
    find all image positions
    @return None if not found else a tuple: (origin.shape, query.shape, postions)
    might raise Exception
    '''
    img1 = cv2.imread(query, 0) # query image(small)
    img2 = cv2.imread(origin, 0) # train image(big)

    # Initiate SIFT detector
    sift = cv2.SIFT()

    # find the keypoints and descriptors with SIFT
    kp1, des1 = sift.detectAndCompute(img1,None)
    kp2, des2 = sift.detectAndCompute(img2,None)
    print len(kp1), len(kp2)

    FLANN_INDEX_KDTREE = 0
    index_params = dict(algorithm = FLANN_INDEX_KDTREE, trees = 5)
    search_params = dict(checks = 50)

    # flann
    flann = cv2.FlannBasedMatcher(index_params, search_params)
    matches = flann.knnMatch(des1, des2, k=2)

    # store all the good matches as per Lowe's ratio test.
    good = []
    for m,n in matches:
        if m.distance < 0.7*n.distance:
            good.append(m)
    print len(kp1), len(kp2), 'good cnt:', len(good)

    if len(good)*1.0/len(kp1) < 0.5:
    #if len(good)<MIN_MATCH_COUNT:
        print "Not enough matches are found - %d/%d" % (len(good),MIN_MATCH_COUNT)
        return img2.shape, img1.shape, []

    queryPts = []
    trainPts = []
    for dm in good:
        queryPts.append(kp1[dm.queryIdx])
        trainPts.append(kp2[dm.trainIdx])

    img3 = cv2.drawKeypoints(img1, queryPts)
    cv2.imwrite('image/query.png', img3)

    img3 = cv2.drawKeypoints(img2, trainPts)
    point = _middlePoint(trainPts)
    print 'position in', point

    if outfile:
        edge = 10
        top_left = (point[0]-edge, point[1]-edge)
        bottom_right = (point[0]+edge, point[1]+edge)
        cv2.rectangle(img3, top_left, bottom_right, 255, 2)
        cv2.imwrite(outfile, img3)
    return img2.shape, img1.shape, [point]

if __name__ == '__main__':
    pts = find_image_position('image/mule.png', 'image/template.png', 
        outfile='image/train.png')
    print pts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Reference
https://github.com/gevent/gevent/blob/master/src/gevent/monkey.py
"""

from __future__ import absolute_import
import sys

if sys.version_info[0] >= 3:
    string_types = str,
    PY3 = True
else:
    import __builtin__ # pylint:disable=import-error
    string_types = __builtin__.basestring,
    PY3 = False


# maps module name -> {attribute name: original item}
# e.g. "time" -> {"sleep": built-in function sleep}
saved = {}


def is_module_patched(modname):
    """Check if a module has been replaced with a cooperative version."""
    return modname in saved


def is_object_patched(modname, objname):
    """Check if an object in a module has been replaced with a cooperative version."""
    return is_module_patched(modname) and objname in saved[modname]


def _get_original(name, items):
    d = saved.get(name, {})
    values = []
    module = None
    for item in items:
        if item in d:
            values.append(d[item])
        else:
            if module is None:
                module = __import__(name)
            values.append(getattr(module, item))
    return values


def get_original(mod_name, item_name):
    """Retrieve the original object from a module.
    If the object has not been patched, then that object will still be retrieved.
    :param item_name: A string or sequence of strings naming the attribute(s) on the module
        ``mod_name`` to return.
    :return: The original value if a string was given for ``item_name`` or a sequence
        of original values if a sequence was passed.
    """
    if isinstance(item_name, string_types):
        return _get_original(mod_name, [item_name])[0]
    else:
        return _get_original(mod_name, item_name)

_NONE = object()


def patch_item(module, attr, newitem):
    olditem = getattr(module, attr, _NONE)
    if olditem is not _NONE:
        saved.setdefault(module, {}).setdefault(attr, olditem)
    setattr(module, attr, newitem)


def remove_item(module, attr):
    olditem = getattr(module, attr, _NONE)
    if olditem is _NONE:
        return
    saved.setdefault(module, {}).setdefault(attr, olditem)
    delattr(module, attr) 


#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Error(Exception):
    def __init__(self, message, data=None):
        self.message = message
        self.data = data

    def __str__(self):
        if self.data:
            return '{}, data: {}'.format(self.message, self.data)
        return self.message

    def __repr__(self):
        return repr(self.message)


class WindowsAppNotFoundError(Error):
    pass


class ImageNotFoundError(Error):
    pass

class WatchTimeoutError(Error):
    pass

class AssertError(Error):
    pass

class AssertExistsError(AssertError):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import base64
import json
import os
import time
import urlparse
from collections import namedtuple

import requests
from PIL import Image
from StringIO import StringIO

from atx.drivers.mixin import DeviceMixin, hook_wrap
from atx.drivers import Display
from atx import consts


DEBUG = False

def convert(dictionary):
    """
    Convert dict to namedtuple
    """
    return namedtuple('GenericDict', dictionary.keys())(**dictionary)


def urljoin(*urls):
    """
    The default urlparse.urljoin behavior look strange
    Standard urlparse.urljoin('http://a.com/foo', '/bar')
    Expect: http://a.com/foo/bar
    Actually: http://a.com/bar

    This function fix that.
    """
    return reduce(urlparse.urljoin, [u.strip('/')+'/' for u in urls if u.strip('/')], '').rstrip('/')


class WebDriverError(Exception):
    def __init__(self, status, value):
        self.status = status
        self.value = value

    def __str__(self):
        return 'WebDriverError(status=%d, value=%s)' % (self.status, self.value)


def httpdo(method, url, data=None):
    """
    Do HTTP Request
    """
    if isinstance(data, dict):
        data = json.dumps(data)
    if DEBUG:
        print "Shell: curl -X {method} -d '{data}' '{url}'".format(method=method, data=data or '', url=url)

    fn = dict(GET=requests.get, POST=requests.post, DELETE=requests.delete)[method]
    response = fn(url, data=data)
    retjson = response.json()
    if DEBUG:
        print 'Return:', json.dumps(retjson, indent=4)
    r = convert(retjson)
    if r.status != 0:
        raise WebDriverError(r.status, r.value)
    return r


class _Client(object):
    def __init__(self, device_url):
        self.__device_url = device_url

    def screenshot(self):
        """Take screenshot
        Return:
            PIL.Image
        """
        url = urljoin(self.__device_url, "screenshot")
        r = httpdo('GET', url)
        raw_image = base64.b64decode(r.value)
        return Image.open(StringIO(raw_image))


class WebDriver(DeviceMixin):
    def __init__(self, device_url):
        DeviceMixin.__init__(self)
        self.__display = None
        self._ymc = _Client(device_url)

    @hook_wrap(consts.EVENT_SCREENSHOT)
    def screenshot(self, filename=None):
        """ Take a screenshot """
        # screen size: 1280x720
        screen = self._ymc.screenshot()
        if filename:
            screen.save(filename)
        return screen

    def start_app(self, bundle_id):
        pass

    def click(self, x, y):
        pass



#!/usr/bin/env python
# coding: utf-8

from androguard.core.bytecodes import apk


def parse_apk(filename):
    '''
    Returns:
        (package, activity)
    '''
    a = apk.APK(filename)
    
    return (a.get_package(), a.get_main_activity())


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Blueprint, not finished yet.

>>>
>>> client = adb.Client(host='127.0.0.1', port=5037)
>>> client.devices()
# {'EP7333W7XB': 'device'}
>>> client.version()
# 1.0.32
>>> client.connect('10.0.0.1', 5555)
True
>>> client.disconnect('10.0.0.1:5555')
True
>>>
>>> dev = client.device(serial=None)
>>> dev.packages()
[{'name': 'com.example.demo', 'version': 2}]
>>> dev.forward_list()
[{'local': 'tcp:8001', 'remote': 'tcp:8000'}]
>>> dev.properties()
{'ro.build.brand', 'MI2', ...}
>>> dev.install('demo.apk')
True
>>> dev.uninstall('com.example.demo', keep_data=True)
True
>>> dev.logcat() # TODO
>>> dev.pull('/data/local/tmp/_screen.png', './')
True
>>> dev.push('./demo.apk', '/data/local/tmp/demo.apk')
True
>>> dev.listdir('/data/local/tmp')
['_screen.png']
>>> dev.shell('ls', '-l', '/data/local/tmp/')
:output as string, replace \r\n to '\n'
>>> dev.start_activity('com.example.demo', '.Client')
None
>>> dev.stat('/data/local/tmp/_screen.png')
:posix.stat_result object
>>> dev.current_app()
com.example.demo
>>> dev.orientation()
: one of [1-4]
>>> dev.screenshot()
: PIL image object
>>> dev.keyevent('HOME')
None
>>> dev.open_minicap()
True
>>> dev.open_minitouch()
True
>>> dev.touch(100, 100)
None
>>> dev.swipe() # TODO
>>> dev.pinch() # only in minitouch
"""

import os
import re
import socket
import subprocess32 as subprocess


LOCAL_PORT = 10300
_init_local_port = LOCAL_PORT - 1

def next_local_port(adb_host=None):
    """ find avaliable free port """
    def is_port_listening(port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((str(adb_host) if adb_host else '127.0.0.1', port))
        s.close()
        return result == 0
    global _init_local_port
    _init_local_port = _init_local_port + 1 if _init_local_port < 32764 else LOCAL_PORT
    while is_port_listening(_init_local_port):
        _init_local_port += 1
    return _init_local_port


class Adb(object):
    __adb_cmd = None

    def __init__(self, serial=None, server_host=None, server_port=None):
        """
        Args:
            - serial: device serial number
            - server_host: adb server host, default 127.0.0.1
            - server_port: adb server port, default 5037
        """
        self.default_serial = serial if serial else os.environ.get("ANDROID_SERIAL", None)
        self.server_host = str(server_host if server_host else '127.0.0.1')
        self.server_port = str(server_port if server_port else '5037')
        self.adb_host_port_options = []
        if self.server_host not in ['localhost', '127.0.0.1']:
            self.adb_host_port_options += ["-H", self.server_host]
        if self.server_port != '5037':
            self.adb_host_port_options += ["-P", self.server_port]
        self.serial = serial or self.device_serial()

    @classmethod
    def adb(cls):
        """return adb binary full path"""
        if cls.__adb_cmd is None:
            if "ANDROID_HOME" in os.environ:
                filename = "adb.exe" if os.name == 'nt' else "adb"
                adb_cmd = os.path.join(os.environ["ANDROID_HOME"], "platform-tools", filename)
                if not os.path.exists(adb_cmd):
                    raise EnvironmentError(
                        "Adb not found in $ANDROID_HOME path: %s." % os.environ["ANDROID_HOME"])
            else:
                import distutils
                if "spawn" not in dir(distutils):
                    import distutils.spawn
                adb_cmd = distutils.spawn.find_executable("adb")
                if adb_cmd:
                    adb_cmd = os.path.realpath(adb_cmd)
                else:
                    raise EnvironmentError("$ANDROID_HOME environment not set.")
            cls.__adb_cmd = adb_cmd
        return cls.__adb_cmd

    def devices(self):
        '''get a dict of attached devices. key is the device serial, value is device name.'''
        out = self.raw_cmd('devices').communicate()[0].decode("utf-8")
        match = "List of devices attached"
        index = out.find(match)
        if index < 0:
            raise EnvironmentError("adb is not working.")
        return dict([s.split("\t") for s in out[index + len(match):].strip().splitlines() 
                if s.strip() and not s.strip().startswith('*')])

    def cmd(self, *args, **kwargs):
        '''adb command, add -s serial by default. return the subprocess.Popen object.'''
        serial = self.device_serial() # TODO(ssx): useless here, need to remove and test
        if serial:
            if " " in serial:  # TODO how to include special chars on command line
                serial = "'%s'" % serial
            return self.raw_cmd(*["-s", serial] + list(args))
        else:
            return self.raw_cmd(*args)

    def raw_cmd(self, *args):
        '''adb command. return the subprocess.Popen object.'''
        cmd_line = [self.adb()] + self.adb_host_port_options + list(args)
        if os.name != "nt":
            cmd_line = [" ".join(cmd_line)]
        return subprocess.Popen(cmd_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def build_cmd(self, *args):
        '''adb command array
        For example:
            build_cmd("shell", "uptime") will get
            ["adb", "-s", "xx..", "-P", "5037", "shell", "uptime"]
        '''
        return [self.adb(), "-s", self.serial] + self.adb_host_port_options + list(args)

    def device_serial(self):
        devices = self.devices()
        if not self.default_serial:
            if devices:
                if len(devices) is 1:
                    self.default_serial = list(devices.keys())[0]
                else:
                    raise EnvironmentError("Multiple devices attached but default android serial not set.")
            else:
                raise EnvironmentError("Device not attached.")
        else:
            if self.default_serial not in devices:
                raise EnvironmentError("Device(%s) not attached." % self.default_serial)

        if devices[self.default_serial] != 'device':
            raise EnvironmentError("Device(%s) is not ready. status(%s)." % 
                (self.default_serial, devices[self.default_serial]))
        return self.default_serial

    def forward(self, device_port, local_port=None):
        '''adb port forward. return local_port'''
        if local_port is None:
            for s, lp, rp in self.forward_list():
                if s == self.device_serial() and rp == 'tcp:%d' % device_port:
                    return int(lp[4:])
            return self.forward(device_port, next_local_port(self.server_host))
        else:
            self.cmd("forward", "tcp:%d" % local_port, "tcp:%d" % device_port).wait()
            return local_port

    def forward_list(self):
        '''adb forward --list'''
        version = self.version()
        if int(version[1]) <= 1 and int(version[2]) <= 0 and int(version[3]) < 31:
            raise EnvironmentError("Low adb version.")
        lines = self.raw_cmd("forward", "--list").communicate()[0].decode("utf-8").strip().splitlines()
        return [line.strip().split() for line in lines]

    def version(self):
        '''adb version'''
        match = re.search(r"(\d+)\.(\d+)\.(\d+)", self.raw_cmd("version").communicate()[0].decode("utf-8"))
        return [match.group(i) for i in range(4)]

    def remove(self, path):
        """Remove remote file
        Return:
            bool: true or false"""
        p = self.cmd('shell', 'rm', path)
        stdout, stderr = p.communicate()
        if stdout or stderr:
            return False
        else:
            return True


if __name__ == '__main__':
    adb = Adb()
    print adb.version()
    print adb.devices()
    print adb.forward_list()
    print adb.forward(8001)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import atexit
import argparse
import os
import time
import json
import warnings
import inspect
import codecs

import imageio

from atx import consts
from atx import errors
from atx import imutils
from atx.base import nameddict
from atx.ext.report import patch as pt
from PIL import Image


__dir__ = os.path.dirname(os.path.abspath(__file__))


class ExtDeprecationWarning(DeprecationWarning):
    pass

warnings.simplefilter('always', ExtDeprecationWarning)

def json2obj(data):
    data['this'] = data.pop('self', None)
    return nameddict('X', data.keys())(**data)

def center(bounds):
    x = (bounds['left'] + bounds['right'])/2
    y = (bounds['top'] + bounds['bottom'])/2
    return (x, y)


class Report(object):
    """
    Example usage:
    from atx.ext.report import Report

    Report(d)
    """
    def __init__(self, d, save_dir='report'):
        image_dir = os.path.join(save_dir, 'images')
        if not os.path.exists(image_dir):
            os.makedirs(image_dir)

        self.d = d
        self.save_dir = save_dir
        self.steps = []
        self.result = None

        self.__gif_path = os.path.join(save_dir, 'output.gif')
        self.__gif = imageio.get_writer(self.__gif_path, fps=2)
        self.__uia_last_position = None
        self.__last_screenshot = None
        self.__closed = False
        
        self.start_record()

    @property
    def last_screenshot(self):
        return self.__last_screenshot

    def _uia_listener(self, evtjson):
        evt = json2obj(evtjson)
        if evt.name != '_click':
            return
        if evt.is_before:
            self.d.screenshot()
            self.__uia_last_position = center(evt.this.bounds)
        else:
            (x, y) = self.__uia_last_position
            # self.last_screenshot
            cv_last_img = imutils.from_pillow(self.last_screenshot)
            cv_last_img = imutils.mark_point(cv_last_img, x, y)
            screen = imutils.to_pillow(cv_last_img)
            screen_before = self._save_screenshot(screen=screen, append_gif=True)
            # FIXME: maybe need sleep for a while
            screen_after = self._save_screenshot(append_gif=True)

            self.add_step('click',
                screen_before=screen_before,
                screen_after=screen_after,
                position={'x': x, 'y': y})

    def add_step(self, action, **kwargs):
        kwargs['success'] = kwargs.pop('success', True)
        kwargs['description'] = kwargs.get('description') or kwargs.get('desc')
        kwargs['time'] = round(kwargs.pop('time', time.time()-self.start_time), 1)
        kwargs['action'] = action
        self.steps.append(kwargs)

    def patch_uiautomator(self):
        """
        Record steps of uiautomator
        """
        import uiautomator
        uiautomator.add_listener('atx-report', self._uia_listener)

    def patch_wda(self):
        """
        Record steps of WebDriverAgent
        """
        import wda

        def _click(that):
            rawx, rawy = that.bounds.center
            x, y = self.d.scale*rawx, self.d.scale*rawy
            screen_before = self._save_screenshot()
            orig_click = pt.get_original(wda.Selector, 'click')
            screen_after = self._save_screenshot()
            self.add_step('click',
                screen_before=screen_before,
                screen_after=screen_after,
                position={'x': x, 'y': y})
            return orig_click(that)

        pt.patch_item(wda.Selector, 'click', _click)

    def start_record(self):
        self.start_time = time.time()

        w, h = self.d.display
        if self.d.rotation in (1, 3): # for horizontal
            w, h = h, w
        self.result = dict(device=dict(
            display=dict(width=w, height=h),
            serial=getattr(self.d, 'serial', ''),
            start_time=time.strftime("%Y-%m-%d %H:%M:%S"),
            start_timestamp=time.time(),
        ), steps=self.steps)

        self.d.add_listener(self._listener, consts.EVENT_ALL) # ^ consts.EVENT_SCREENSHOT)

        self.__closed = False
        atexit.register(self.close)

    def close(self):
        if self.__closed:
            return

        save_dir = self.save_dir
        data = json.dumps(self.result)
        tmpl_path = os.path.join(__dir__, 'index.tmpl.html')
        save_path = os.path.join(save_dir, 'index.html')
        json_path = os.path.join(save_dir, 'result.json')

        with codecs.open(tmpl_path, 'rb', 'utf-8') as f:
            html_content = f.read().replace('$$data$$', data)

        with open(json_path, 'wb') as f:
            f.write(json.dumps(self.result, indent=4).encode('utf-8'))

        with open(save_path, 'wb') as f:
            f.write(html_content.encode('utf-8'))

        self.__gif.close()
        self.__closed = True

    def info(self, text, screenshot=None):
        """
        Args:
            - text(str): description
            - screenshot: Bool or PIL.Image object
        """
        step = {
            'time': '%.1f' % (time.time()-self.start_time,),
            'action': 'info',
            'description': text,
            'success': True,
        }   
        if screenshot:
            step['screenshot'] = self._take_screenshot(screenshot, name_prefix='info')
        self.steps.append(step)

    def error(self, text, screenshot=None):
        """
        Args:
            - text(str): description
            - screenshot: Bool or PIL.Image object
        """
        step = {
            'time': '%.1f' % (time.time()-self.start_time,),
            'action': 'error',
            'description': text,
            'success': False,
        }   
        if screenshot:
            step['screenshot'] = self._take_screenshot(screenshot, name_prefix='error')
        self.steps.append(step)

    def _save_screenshot(self, screen=None, name=None, name_prefix='screen', append_gif=False):
        if screen is None:
            screen = self.d.screenshot()
        if name is None:
            name = 'images/%s_%d.jpg' % (name_prefix, time.time())
        relpath = os.path.join(self.save_dir, name)
        screen.save(relpath)
        if append_gif:
            self._add_to_gif(screen)
        return name

    def _add_to_gif(self, image):
        half = 0.5
        out = image.resize([int(half*s) for s in image.size])
        cvimg = imutils.from_pillow(out)
        self.__gif.append_data(cvimg[:, :, ::-1])

    def _take_screenshot(self, screenshot=False, name_prefix='unknown'):
        """
        This is different from _save_screenshot.
        The return value maybe None or the screenshot path

        Args:
            screenshot: bool or PIL image
        """
        if isinstance(screenshot, bool):
            if not screenshot:
                return
            return self._save_screenshot(name_prefix=name_prefix)
        if isinstance(screenshot, Image.Image):
            return self._save_screenshot(screen=screenshot, name_prefix=name_prefix)

        raise TypeError("invalid type for func _take_screenshot: "+ type(screenshot))

    def _record_assert(self, is_success, text, screenshot=False, desc=None):
        step = {
            'time': '%.1f' % (time.time()-self.start_time,),
            'action': 'assert',
            'message': text,
            'description': desc,
            'success': is_success,
            'screenshot': self._take_screenshot(screenshot, name_prefix='assert'),
        }
        self.steps.append(step)

    def _add_assert(self, **kwargs):
        """
        if screenshot is None, only failed case will take screenshot
        """
        # convert screenshot to relative path from <None|True|False|PIL.Image>
        screenshot = kwargs.get('screenshot')
        is_success = kwargs.get('success')
        screenshot = (not is_success) if screenshot is None else screenshot
        kwargs['screenshot'] = self._take_screenshot(screenshot=screenshot, name_prefix='assert')
        self.add_step('assert', **kwargs)
        if not is_success:
            message = kwargs.get('message')
            frame, filename, line_number, function_name, lines, index = inspect.stack()[2]
            print('Assert [%s: %d] WARN: %s' % (filename, line_number, message))
            if not kwargs.get('safe', False):
                raise AssertionError(message)

    def assert_equal(self, v1, v2, **kwargs):#, desc=None, screenshot=False, safe=False):
        """ Check v1 is equals v2, and take screenshot if not equals
        Args:
            - desc (str): some description
            - safe (bool): will omit AssertionError if set to True
            - screenshot: can be type <None|True|False|PIL.Image>
        """
        is_success = v1 == v2
        if is_success:
            message = "assert equal success, %s == %s" %(v1, v2)
        else:
            message = '%s not equal %s' % (v1, v2)
        kwargs.update({
            'message': message,
            'success': is_success,
        })
        self._add_assert(**kwargs)

    def assert_image_exists(self, pattern, timeout=20.0, **kwargs):
        """TODO: not finished yet.
        Assert if image exists
        Args:
            - pattern: image filename # not support pattern for now
            - timeout (float): seconds
        """
        start_time = time.time()
        is_success = False
        while time.time() - start_time < timeout:
            point = self.match(pattern, **match_kwargs)
            if point is None:
                continue
            if not point.matched:
                # log.debug('Ignore confidence: %s', point.confidence)
                continue
            is_success = True
        kwargs.update({
            'message': 'image exists',
            'success': is_success,
        })
        self._add_assert(**kwargs)

    def assert_ui_exists(self, ui, **kwargs):
        """ For Android & IOS
        Args:
            - ui: need have property "exists"
            - desc (str): description
            - safe (bool): will omit AssertionError if set to True
            - screenshot: can be type <None|True|False|PIL.Image>
            - platform (str, default:android): android | ios
        """
        is_success = ui.exists
        if is_success:
            if kwargs.get('screenshot') is not None:
                if self.d.platform == 'android':
                    bounds = ui.info['bounds'] # For android only.
                    kwargs['position'] = {
                        'x': (bounds['left']+bounds['right'])/2,
                        'y': (bounds['top']+bounds['bottom'])/2,
                    }
                elif self.d.platform == 'ios':
                    bounds = ui.bounds # For iOS only.
                    kwargs['position'] = {
                        'x': self.d.scale*(bounds.x+bounds.width/2),
                        'y': self.d.scale*(bounds.y+bounds.height/2),
                    }
            message = 'UI exists'
        else:
            message = 'UI not exists'
        kwargs.update({
            'message': message,
            'success': is_success,
        })
        self._add_assert(**kwargs)

    def _listener(self, evt):
        d = self.d
        screen_before = 'images/before_%d.jpg' % time.time()
        screen_before_abspath = os.path.join(self.save_dir, screen_before)

        # keep screenshot for every call
        if not evt.is_before and evt.flag == consts.EVENT_SCREENSHOT:
            self.__last_screenshot = evt.retval

        if evt.depth > 1: # base depth is 1
            return

        if evt.is_before: # call before function
            if evt.flag == consts.EVENT_CLICK:
                self.__last_screenshot = d.screenshot() # Maybe no need to set value here.
                (x, y) = evt.args
                cv_img = imutils.from_pillow(self.last_screenshot)
                cv_img = imutils.mark_point(cv_img, x, y)
                self.__last_screenshot = imutils.to_pillow(cv_img)
                self._add_to_gif(self.last_screenshot)
            return

        if evt.flag == consts.EVENT_CLICK:
            if self.last_screenshot: # just in case
                self.last_screenshot.save(screen_before_abspath)
            screen_after = 'images/after_%d.jpg' % time.time()
            d.screenshot(os.path.join(self.save_dir, screen_after))

            (x, y) = evt.args
            self.add_step('click',
                screen_before=screen_before,
                screen_after=screen_after,
                position={'x': x, 'y': y})
        elif evt.flag == consts.EVENT_CLICK_IMAGE:
            kwargs = {
                'success': evt.traceback is None,
                'traceback': None if evt.traceback is None else evt.traceback.stack,
                'description': evt.kwargs.get('desc'),
            }
            # not record if image not found
            if evt.retval is None and evt.traceback is None:
                return
            
            if self.last_screenshot:
                self.last_screenshot.save(screen_before_abspath)
                kwargs['screen_before'] = screen_before
            if evt.traceback is None or not isinstance(evt.traceback.exception, IOError):
                target = 'images/target_%d.jpg' % time.time()
                pattern = d.pattern_open(evt.args[0])
                self._save_screenshot(pattern, name=target)
                kwargs['target'] = target
            if evt.traceback is None:
                (x, y) = evt.retval.pos
                # FIXME(ssx): quick hot fix
                cv_img = imutils.from_pillow(self.last_screenshot)
                cv_img = imutils.mark_point(cv_img, x, y)
                self.__last_screenshot = imutils.to_pillow(cv_img)
                self.last_screenshot.save(screen_before_abspath)
                
                screen_after = 'images/after_%d.jpg' % time.time()
                d.screenshot(os.path.join(self.save_dir, screen_after))
                kwargs['screen_after'] = screen_after
                kwargs['confidence'] = evt.retval.confidence
                kwargs['position'] = {'x': x, 'y': y}
            self.add_step('click_image', **kwargs)
        elif evt.flag == consts.EVENT_ASSERT_EXISTS: # this is image, not tested
            pattern = d.pattern_open(evt.args[0])
            target = 'images/target_%.2f.jpg' % time.time()
            self._save_screenshot(pattern, name=target)
            kwargs = {
                'target': target,
                'description': evt.kwargs.get('desc'),
                'screen': self._save_screenshot(name='images/screen_%.2f.jpg' % time.time()),
                'traceback': None if evt.traceback is None else evt.traceback.stack,
                'success': evt.traceback is None,
            }
            if evt.traceback is None:
                kwargs['confidence'] = evt.retval.confidence
                (x, y) = evt.retval.pos
                kwargs['position'] = {'x': x, 'y': y}
            self.add_step('assert_exists', **kwargs)


def listen(d, save_dir='report'):
    ''' Depreciated '''
    warnings.warn(
        "Using report.listen is deprecated, use report.Report(d, save_dir) instead.", 
        ExtDeprecationWarning, stacklevel=2
    )
    Report(d, save_dir)



# coding: utf-8
import os
import atx


__basename = os.path.basename(os.path.splitext(__file__)[0])
d = atx.connect(platform="android")
d.image_path = [".", "images", os.path.join("images", __basename)]

d.screenshot('screen.png')
if 0 == 0:
  print('你好')
  for count in range(1):
    print('Hello world')
print('Hello world')


#!/usr/bin/env python
# coding: utf-8
#
# Usage: python -matx screen [-s 0.8]
from __future__ import absolute_import

import time
from cStringIO import StringIO
from PIL import Image

from atx import adbkit


def main(host=None, port=None, serial=None, scale=1.0, out='screenshot.png', method='minicap'):
    """
    If minicap not avaliable then use uiautomator instead

    Disable scale for now.
    Because -s scale is conflict of -s serial
    """
    print('Started screencap')
    start = time.time()

    client = adbkit.Client(host=host, port=port)
    device = client.device(serial)
    im = device.screenshot(scale=scale)
    im.save(out)
    print('Time spend: %.2fs' % (time.time() - start))
    print('File saved to "%s"' % out)

    try:
        import win32clipboard

        output = StringIO()
        im.convert("RGB").save(output, "BMP")
        data = output.getvalue()[14:]
        output.close()

        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(win32clipboard.CF_DIB, data)
        win32clipboard.CloseClipboard()
        print('Copied to clipboard')
    except:
        pass # ignore


if __name__ == '__main__':
    main()



#-*- encoding: utf-8 -*-

import os
import os.path
import time

from atx.record.android import RecordDevice, AndroidRecorder
from atx.record.draft_editor import run as run_draft_editor

def main(serial=None, host=None, port=None, workdir=".", nonui_activities=None, edit_mode=False):
    workdir = os.path.abspath(workdir)
    
    if edit_mode:
        run_draft_editor(workdir, None)
        return

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    d = RecordDevice(serialno=serial, host=host, port=port)

    rec = AndroidRecorder(d, workdir)
    if nonui_activities:
        for a in nonui_activities:
            rec.add_nonui_activity(a)

    rec.start()

    try:
        time.sleep(4)
        print '-'*20 + ' STARTED ' + '-'*20
        print 'Please operate on the phone. Press Ctrl+C to stop.'
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        pass
        
    rec.stop()
    print '-'*20 + ' STOPPED ' + '-'*20

    if len(rec.frames) > 0:
        print 'start web service to modify recorded case'
        run_draft_editor(workdir, None)
    else:
        print 'No action recorded.'

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from atx import base
from atx import strutils


def test_list_all_image():
    images = base.list_all_image('media')
    images = list(images)
    assert len(images) > 0
    #print list(images)
    #assert 'media/haima.png' in images or 'media\\haima.png' in images


def test_image_name_match():
    match = base.image_name_match
    assert match('foo.png', 'foo.png')
    assert match('./foo.png', 'foo.png')
    assert match('.//foo.png', './foo.png')
    assert match('foo', 'foo.png')
    assert match('foo', 'foo.PNG')
    assert match('foo', 'foo.jpg')
    assert match('foo', 'bar/foo.png')
    assert match('foo', 'bar/./foo.png')
    assert match('foo', 'foo@rsl(1280x720).png')
    assert match('foo', 'bar/foo@rsl(1280x720).png')
    assert match('./foo', 'bar/foo@rsl(1280x720).png')
    assert not match('foo.txt', 'foo.png')
    assert not match('foo.jpg', 'foo.png')
    assert not match('bar/foo', 'foo.jpg')


def test_search_image():
    imgpath = base.search_image('haima', path=['media'])
    assert imgpath is not None
    assert strutils.encode('haima.png') in imgpath
    assert strutils.encode('media') in imgpath


def test_filename_match():
    # @auto is supported
    assert base.filename_match('fight@auto.png', 'fight@4x3.png', 4, 3) == True
    assert base.filename_match('fight@auto.png', 'fight@4x3.png', 3, 4) == True
    assert base.filename_match('fight@auto.png', 'fight.4x3.png', 3, 4) == True
    assert base.filename_match('fight@auto.png', 'fight.3x4.png', 3, 4) == True

    assert base.filename_match('fight@auto.png', 'fight.16x9.png', 3, 4) == False
    assert base.filename_match('fight@auto.png', 'fight@16x9.png', 0, 0) == False

    # not support @wxh now
    assert base.filename_match('fight@3x4.png', 'fight@3x4.png', 3, 4) == True

    # normal
    assert base.filename_match('fight.png', 'fight.png', 3, 4) == True
    assert base.filename_match('fight.png', 'fight.png', 0, 0) == True
    assert base.filename_match('fight.png', 'fight.jpg', 0, 0) == False
    # assert base.filename_match('fight.3x4.png', 'fight.3x4.png', 3, 4) == True
    # assert base.filename_match('fight@auto.png', 'fight@4x3.png', 15, 20) == False


if __name__ == '__main__':
    test_filename_match()
